import multiprocessing
import os
import sys
import threading
from collections.abc import Iterator, Mapping, Sequence
from contextlib import ExitStack
from multiprocessing.context import BaseContext as MultiprocessingBaseContext
from multiprocessing.process import BaseProcess
from typing import TYPE_CHECKING, Any, Optional

from dagster import _check as check
from dagster._core.definitions.metadata import MetadataValue
from dagster._core.definitions.reconstruct import ReconstructableJob
from dagster._core.definitions.repository_definition import RepositoryLoadData
from dagster._core.errors import (
    DagsterExecutionInterruptedError,
    DagsterSubprocessError,
    DagsterUnmetExecutorRequirementsError,
)
from dagster._core.events import DagsterEvent, EngineEventData
from dagster._core.execution.api import create_execution_plan, execute_plan_iterator
from dagster._core.execution.context.system import IStepContext, PlanOrchestrationContext
from dagster._core.execution.context_creation_job import create_context_free_log_manager
from dagster._core.execution.plan.active import ActiveExecution
from dagster._core.execution.plan.instance_concurrency_context import InstanceConcurrencyContext
from dagster._core.execution.plan.plan import ExecutionPlan
from dagster._core.execution.plan.state import KnownExecutionState
from dagster._core.execution.plan.step import ExecutionStep
from dagster._core.execution.retries import RetryMode
from dagster._core.executor.base import Executor
from dagster._core.executor.child_process_executor import (
    ChildProcessCommand,
    ChildProcessCrashException,
    ChildProcessEvent,
    ChildProcessSystemErrorEvent,
    execute_child_process_command,
)
from dagster._core.instance import DagsterInstance
from dagster._utils import get_run_crash_explanation, start_termination_thread
from dagster._utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info
from dagster._utils.timing import TimerResult, format_duration, time_execution_scope

if TYPE_CHECKING:
    from dagster._core.instance.ref import InstanceRef
    from dagster._core.storage.dagster_run import DagsterRun

DELEGATE_MARKER = "multiprocess_subprocess_init"


class MultiprocessExecutorChildProcessCommand(ChildProcessCommand):
    def __init__(
        self,
        run_config: Mapping[str, object],
        dagster_run: "DagsterRun",
        step_key: str,
        instance_ref: "InstanceRef",
        term_event: Any,
        recon_pipeline: ReconstructableJob,
        retry_mode: RetryMode,
        known_state: Optional[KnownExecutionState],
        repository_load_data: Optional[RepositoryLoadData],
    ):
        self.run_config = run_config
        self.dagster_run = dagster_run
        self.step_key = step_key
        self.instance_ref = instance_ref
        self.term_event = term_event
        self.recon_pipeline = recon_pipeline
        self.retry_mode = retry_mode
        self.known_state = known_state
        self.repository_load_data = repository_load_data

    def execute(self) -> Iterator[DagsterEvent]:
        recon_job = self.recon_pipeline
        with DagsterInstance.from_ref(self.instance_ref) as instance:
            done_event = threading.Event()
            start_termination_thread(self.term_event, done_event)
            try:
                log_manager = create_context_free_log_manager(instance, self.dagster_run)

                yield DagsterEvent.step_worker_started(
                    log_manager,
                    self.dagster_run.job_name,
                    message=f'Executing step "{self.step_key}" in subprocess.',
                    metadata={
                        "pid": MetadataValue.text(str(os.getpid())),
                    },
                    step_key=self.step_key,
                )
                execution_plan = create_execution_plan(
                    job=recon_job,
                    run_config=self.run_config,
                    step_keys_to_execute=[self.step_key],
                    known_state=self.known_state,
                    repository_load_data=self.repository_load_data,
                )
                yield from execute_plan_iterator(
                    execution_plan,
                    recon_job,
                    self.dagster_run,
                    run_config=self.run_config,
                    retry_mode=self.retry_mode.for_inner_plan(),
                    instance=instance,
                )
            finally:
                # set events to stop the termination thread on exit
                done_event.set()  # waiting on term_event so set done first
                self.term_event.set()


class MultiprocessExecutor(Executor):
    def __init__(
        self,
        retries: RetryMode,
        max_concurrent: Optional[int],
        tag_concurrency_limits: Optional[list[dict[str, Any]]] = None,
        start_method: Optional[str] = None,
        explicit_forkserver_preload: Optional[Sequence[str]] = None,
    ):
        self._retries = check.inst_param(retries, "retries", RetryMode)
        if not max_concurrent:
            env_var_default = os.getenv("DAGSTER_MULTIPROCESS_EXECUTOR_MAX_CONCURRENT")
            max_concurrent = (
                int(env_var_default) if env_var_default else multiprocessing.cpu_count()
            )

        self._max_concurrent = check.int_param(max_concurrent, "max_concurrent")
        self._tag_concurrency_limits = check.opt_list_param(
            tag_concurrency_limits, "tag_concurrency_limits"
        )
        start_method = check.opt_str_param(start_method, "start_method")
        valid_starts = multiprocessing.get_all_start_methods()

        if start_method is None:
            start_method = "spawn"

        if start_method not in valid_starts:
            raise DagsterUnmetExecutorRequirementsError(
                f"The selected start_method '{start_method}' is not available. "
                f"Only {valid_starts} are valid options on {sys.platform} python {sys.version}.",
            )
        self._start_method = start_method
        self._explicit_forkserver_preload = explicit_forkserver_preload

    @property
    def retries(self) -> RetryMode:
        return self._retries

    def execute(
        self, plan_context: PlanOrchestrationContext, execution_plan: ExecutionPlan
    ) -> Iterator[DagsterEvent]:
        check.inst_param(plan_context, "plan_context", PlanOrchestrationContext)
        check.inst_param(execution_plan, "execution_plan", ExecutionPlan)

        job = plan_context.reconstructable_job

        multiproc_ctx = multiprocessing.get_context(self._start_method)
        if self._start_method == "forkserver":
            module = job.get_module()
            # if explicitly listed in config we will use that
            if self._explicit_forkserver_preload is not None:
                preload = self._explicit_forkserver_preload

            # or if the reconstructable job has a module target, we will use that
            elif module is not None:
                preload = [module]

            # base case is to preload the dagster library
            else:
                preload = ["dagster"]

            # we import this module first to avoid user code like
            # pyspark.serializers._hijack_namedtuple from breaking us
            if "dagster._core.executor.multiprocess" not in preload:
                preload = ["dagster._core.executor.multiprocess", *preload]

            multiproc_ctx.set_forkserver_preload(list(preload))

        limit = self._max_concurrent
        tag_concurrency_limits = self._tag_concurrency_limits

        yield DagsterEvent.engine_event(
            plan_context,
            f"Executing steps using multiprocess executor: parent process (pid: {os.getpid()})",
            event_specific_data=EngineEventData.multiprocess(
                os.getpid(), step_keys_to_execute=execution_plan.step_keys_to_execute
            ),
        )

        timer_result: Optional[TimerResult] = None
        with ExitStack() as stack:
            timer_result = stack.enter_context(time_execution_scope())

            instance_concurrency_context = stack.enter_context(
                InstanceConcurrencyContext(plan_context.instance, plan_context.dagster_run)
            )
            active_execution = stack.enter_context(
                ActiveExecution(
                    execution_plan,
                    retry_mode=self.retries,
                    max_concurrent=limit,
                    tag_concurrency_limits=tag_concurrency_limits,
                    instance_concurrency_context=instance_concurrency_context,
                )
            )
            active_iters: dict[str, Iterator[Optional[DagsterEvent]]] = {}
            errors: dict[int, SerializableErrorInfo] = {}
            processes: dict[str, BaseProcess] = {}
            term_events: dict[str, Any] = {}
            stopping: bool = False

            try:
                while (not stopping and not active_execution.is_complete) or active_iters:
                    if active_execution.check_for_interrupts():
                        yield DagsterEvent.engine_event(
                            plan_context,
                            "Multiprocess executor: received termination signal - "
                            "forwarding to active child processes",
                            EngineEventData.interrupted(list(active_iters.keys())),
                        )
                        stopping = True
                        active_execution.mark_interrupted()
                        for key, term_event in term_events.items():
                            if key in processes:
                                if processes[key].is_alive():
                                    term_event.set()
                                del processes[key]

                    while not stopping:
                        steps = active_execution.get_steps_to_execute(
                            limit=(limit - len(active_iters)),
                        )

                        yield from active_execution.concurrency_event_iterator(plan_context)

                        if not steps:
                            break

                        for step in steps:
                            step_context = plan_context.for_step(step)
                            term_events[step.key] = multiproc_ctx.Event()
                            active_iters[step.key] = execute_step_out_of_process(
                                multiproc_ctx,
                                job,
                                step_context,
                                step,
                                errors,
                                processes,
                                term_events,
                                self.retries,
                                active_execution.get_known_state(),
                                execution_plan.repository_load_data,
                            )

                    # process active iterators
                    empty_iters = []
                    for key, step_iter in active_iters.items():
                        try:
                            event_or_none = next(step_iter)
                            if event_or_none is None:
                                continue
                            else:
                                yield event_or_none
                                active_execution.handle_event(event_or_none)

                        except ChildProcessCrashException as crash:
                            serializable_error = serializable_error_info_from_exc_info(
                                sys.exc_info()
                            )
                            step_context = plan_context.for_step(
                                active_execution.get_step_by_key(key)
                            )
                            yield DagsterEvent.engine_event(
                                step_context,
                                get_run_crash_explanation(
                                    prefix=f"Multiprocess executor: child process for step {key}",
                                    exit_code=crash.exit_code,  # pyright: ignore[reportArgumentType]
                                ),
                                EngineEventData.engine_error(serializable_error),
                            )
                            failure_or_retry_event = self.get_failure_or_retry_event_after_crash(
                                step_context, serializable_error, active_execution.get_known_state()
                            )

                            active_execution.handle_event(failure_or_retry_event)
                            yield failure_or_retry_event
                            empty_iters.append(key)
                            errors[crash.pid] = serializable_error
                        except StopIteration:
                            empty_iters.append(key)

                    # clear and mark complete finished iterators
                    for key in empty_iters:
                        del active_iters[key]
                        del term_events[key]
                        if key in processes:
                            del processes[key]
                        active_execution.verify_complete(plan_context, key)

                    # process skipped and abandoned steps
                    yield from active_execution.plan_events_iterator(plan_context)
            except Exception:
                if not stopping and active_iters:
                    serializable_error = serializable_error_info_from_exc_info(sys.exc_info())
                    yield DagsterEvent.engine_event(
                        plan_context,
                        "Unexpected exception while steps were still in-progress - terminating running steps:",
                        event_specific_data=EngineEventData(
                            metadata={
                                "steps_interrupted": MetadataValue.text(
                                    str(list(active_iters.keys()))
                                )
                            },
                            error=serializable_error,
                        ),
                    )
                    for key, term_event in term_events.items():
                        if key in processes:
                            if processes[key].is_alive():
                                term_event.set()
                            del processes[key]

                raise

            errs = {pid: err for pid, err in errors.items() if err}

            # After termination starts, raise an interrupted exception once all subprocesses
            # have finished cleaning up (and the only errors were from being interrupted)
            if (
                stopping
                and (not active_iters)
                and all(
                    [
                        err_info.cls_name
                        in {"DagsterExecutionInterruptedError", "KeyboardInterrupt"}
                        for err_info in errs.values()
                    ]
                )
            ):
                yield DagsterEvent.engine_event(
                    plan_context,
                    "Multiprocess executor: interrupted all active child processes",
                    event_specific_data=EngineEventData(),
                )
                raise DagsterExecutionInterruptedError()
            elif errs:
                raise DagsterSubprocessError(
                    "During multiprocess execution errors occurred in child"
                    " processes:\n{error_list}".format(
                        error_list="\n".join(
                            [f"In process {pid}: {err.to_string()}" for pid, err in errs.items()]
                        )
                    ),
                    subprocess_error_infos=list(errs.values()),
                )

        if timer_result:
            yield DagsterEvent.engine_event(
                plan_context,
                f"Multiprocess executor: parent process exiting after {format_duration(timer_result.millis)} (pid: {os.getpid()})",
                event_specific_data=EngineEventData.multiprocess(os.getpid()),
            )


def execute_step_out_of_process(
    multiproc_ctx: MultiprocessingBaseContext,
    recon_job: ReconstructableJob,
    step_context: IStepContext,
    step: ExecutionStep,
    errors: dict[int, SerializableErrorInfo],
    processes: dict[str, BaseProcess],
    term_events: dict[str, Any],
    retries: RetryMode,
    known_state: KnownExecutionState,
    repository_load_data: Optional[RepositoryLoadData],
) -> Iterator[Optional[DagsterEvent]]:
    command = MultiprocessExecutorChildProcessCommand(
        run_config=step_context.run_config,
        dagster_run=step_context.dagster_run,
        step_key=step.key,
        instance_ref=step_context.instance.get_ref(),
        term_event=term_events[step.key],
        recon_pipeline=recon_job,
        retry_mode=retries,
        known_state=known_state,
        repository_load_data=repository_load_data,
    )

    yield DagsterEvent.step_worker_starting(
        step_context,
        f'Launching subprocess for "{step.key}".',
        metadata={},
    )

    for ret in execute_child_process_command(multiproc_ctx, command):
        if ret is None or isinstance(ret, DagsterEvent):
            yield ret
        elif isinstance(ret, ChildProcessEvent):
            if isinstance(ret, ChildProcessSystemErrorEvent):
                errors[ret.pid] = ret.error_info
        elif isinstance(ret, BaseProcess):
            processes[step.key] = ret
        else:
            check.failed(f"Unexpected return value from child process {type(ret)}")



from dagster._core.definitions.declarative_automation.operands.operands import (
    BackfillInProgressAutomationCondition as BackfillInProgressAutomationCondition,
    CheckResultCondition as CheckResultCondition,
    CodeVersionChangedCondition as CodeVersionChangedCondition,
    CronTickPassedCondition as CronTickPassedCondition,
    ExecutionFailedAutomationCondition as ExecutionFailedAutomationCondition,
    InitialEvaluationCondition as InitialEvaluationCondition,
    InLatestTimeWindowCondition as InLatestTimeWindowCondition,
    LatestRunExecutedWithRootTargetCondition as LatestRunExecutedWithRootTargetCondition,
    MissingAutomationCondition as MissingAutomationCondition,
    NewlyRequestedCondition as NewlyRequestedCondition,
    NewlyUpdatedCondition as NewlyUpdatedCondition,
    RunInProgressAutomationCondition as RunInProgressAutomationCondition,
    WillBeRequestedCondition as WillBeRequestedCondition,
)



import logging
import os
from collections.abc import Mapping
from typing import Any, Optional

from dotenv import dotenv_values
from typing_extensions import Self

from dagster._config import Field, StringSource
from dagster._core.secrets.loader import SecretsLoader
from dagster._serdes import ConfigurableClass
from dagster._serdes.config_class import ConfigurableClassData


def get_env_var_dict(base_dir: str) -> dict[str, str]:
    env_file_path = os.path.join(base_dir, ".env")
    if not os.path.exists(env_file_path):
        return {}

    return {key: val for key, val in dotenv_values(env_file_path).items() if val is not None}


class EnvFileLoader(SecretsLoader, ConfigurableClass):
    def __init__(
        self,
        inst_data: Optional[ConfigurableClassData] = None,
        base_dir=None,
    ):
        self._inst_data = inst_data
        self._base_dir = base_dir or os.getcwd()

    def get_secrets_for_environment(self, location_name: Optional[str]) -> dict[str, str]:
        env_file_path = os.path.join(self._base_dir, ".env")

        env_var_dict = get_env_var_dict(env_file_path)

        if len(env_var_dict):
            logging.getLogger("dagster").info(
                "Loaded environment variables from .env file: "
                + ",".join([env_var for env_var in env_var_dict]),
            )
        else:
            logging.getLogger("dagster").info("No environment variables in .env file")

        return env_var_dict

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {"base_dir": Field(StringSource, is_required=False)}

    @classmethod
    def from_config_value(
        cls, inst_data: ConfigurableClassData, config_value: Mapping[str, Any]
    ) -> Self:
        return cls(inst_data=inst_data, **config_value)



import gc
import logging
import os
import os.path
import threading
from sys import platform
from typing import Optional, Union

import dagster._check as check
from dagster._core.execution.telemetry import RunTelemetryData
from dagster._core.instance import DagsterInstance
from dagster._core.storage.dagster_run import DagsterRun
from dagster._utils.container import (
    UNCONSTRAINED_CGROUP_MEMORY_LIMIT,
    compatible_cgroup_version_detected,
    retrieve_containerized_utilization_metrics,
)

DEFAULT_RUN_METRICS_POLL_INTERVAL_SECONDS = 30.0
DEFAULT_RUN_METRICS_SHUTDOWN_SECONDS = 15


def _get_platform_name() -> str:
    platform_name = platform
    return platform_name.casefold()


def _process_is_containerized() -> bool:
    """Detect if the current process is running in a container under linux."""
    if _get_platform_name() != "linux":
        return False

    # examining the root process (pid==1) under linux to determine if we can rule out being in a container
    file = "/proc/1/exe"
    if not os.path.isfile(file) or not os.access(file, os.R_OK):
        # /proc/1/exe should exist on linux; if it doesn't, we don't know what kind of system we're on
        return False

    target = os.readlink(file)
    if os.path.split(target)[-1] == "init":
        # SysVinit and upstart init - ex: /init, /sbin/init, /etc/init
        if not target == "/dev/init":
            # docker with `--init` will have /dev/init as the target
            return False
    elif os.path.split(target)[-1] == "systemd":
        # the default init process on systemd
        return False

    # can we detect cgroup v1 or v2?
    return compatible_cgroup_version_detected()


def _metric_tags(dagster_run: DagsterRun) -> dict[str, str]:
    location_name = os.getenv("DAGSTER_CLOUD_LOCATION_NAME", None)

    # organization and deployment name are added by the graphql mutation
    tags = {
        "job_name": dagster_run.job_name,
        "location_name": location_name,
        "run_id": dagster_run.run_id,
    }
    # filter out none values
    return {k: v for k, v in tags.items() if v is not None}


def _get_container_metrics(
    previous_cpu_usage_ms: Optional[float] = None,
    previous_measurement_timestamp: Optional[float] = None,
    logger: Optional[logging.Logger] = None,
) -> dict[str, Union[float, None]]:
    metrics = retrieve_containerized_utilization_metrics(logger=logger)

    # calculate cpu_limit
    cpu_quota_us = metrics.get("cpu_cfs_quota_us")
    cpu_period_us = metrics.get("cpu_cfs_period_us")
    cpu_usage = metrics.get("cpu_usage")
    cpu_usage_ms = None
    if cpu_usage is not None:
        cpu_usage_ms = check.not_none(cpu_usage) * 1000  # convert from seconds to milliseconds

    cpu_limit_ms = None
    if cpu_quota_us and cpu_quota_us > 0 and cpu_period_us and cpu_period_us > 0:
        # Why the 1000 factor is a bit counterintuitive:
        # quota / period -> fraction of cpu per unit of time
        # 1000 * quota / period -> ms/sec of cpu
        cpu_limit_ms = (1000.0 * check.not_none(cpu_quota_us)) / check.not_none(cpu_period_us)

    cpu_usage_rate_ms = None
    measurement_timestamp = metrics.get("measurement_timestamp")
    if (previous_cpu_usage_ms and cpu_usage_ms and cpu_usage_ms >= previous_cpu_usage_ms) and (
        previous_measurement_timestamp
        and measurement_timestamp
        and measurement_timestamp > previous_measurement_timestamp
    ):
        cpu_usage_rate_ms = (cpu_usage_ms - previous_cpu_usage_ms) / (
            measurement_timestamp - previous_measurement_timestamp
        )

    cpu_percent = None
    if (cpu_limit_ms and cpu_limit_ms > 0) and (cpu_usage_rate_ms and cpu_usage_rate_ms > 0):
        cpu_percent = 100.0 * check.not_none(cpu_usage_rate_ms) / check.not_none(cpu_limit_ms)

    memory_percent = None
    memory_limit = None
    cgroup_memory_limit = metrics.get("memory_limit")
    memory_usage = metrics.get("memory_usage")
    if (cgroup_memory_limit and 0 < cgroup_memory_limit < UNCONSTRAINED_CGROUP_MEMORY_LIMIT) and (
        memory_usage and memory_usage >= 0
    ):
        memory_limit = cgroup_memory_limit
        memory_percent = 100.0 * check.not_none(memory_usage) / check.not_none(memory_limit)

    return {
        # TODO - eventually we should replace and remove cpu_usage_ms and only send cpu_usage_rate_ms
        #  We need to ensure that the UI and GraphQL queries can handle the change
        #  and that we have 15 days of runs to ensure continuity.
        #  Why? the derivative calculated by datadog on cpu_usage_ms is sometimes wonky, resulting
        #  in a perplexing graph for the end user.
        "container.cpu_usage_ms": cpu_usage_ms,
        "container.cpu_usage_rate_ms": cpu_usage_rate_ms,
        "container.cpu_limit_ms": cpu_limit_ms,
        "container.cpu_percent": cpu_percent,
        "container.memory_usage": memory_usage,
        "container.memory_limit": memory_limit,
        "container.memory_percent": memory_percent,
        "measurement_timestamp": measurement_timestamp,
    }


def _get_python_runtime_metrics() -> dict[str, float]:
    gc_stats = gc.get_stats()

    stats_dict = {}
    for index, gen_dict in enumerate(gc_stats):
        gen_metrics = {
            f"python.runtime.gc_gen_{index}.{key}": value for key, value in gen_dict.items()
        }
        stats_dict.update(gen_metrics)

    return {**stats_dict, "python.runtime.gc_freeze_count": gc.get_freeze_count()}


def _report_run_metrics(
    instance: DagsterInstance,
    dagster_run: DagsterRun,
    metrics: dict[str, float],
    run_tags: dict[str, str],
):
    datapoints: dict[str, float] = {}
    for metric, value in metrics.items():
        if value is None:
            continue
        try:
            datapoints[metric] = float(value)
        except ValueError:
            logging.warning(f"Failed to convert metric value to float: {metric}={value}, skipping")

    telemetry_data = RunTelemetryData(run_id=dagster_run.run_id, datapoints=datapoints)

    # TODO - this should throw an exception or return a control value if the telemetry is not enabled server side
    #   so that we can catch and stop the thread.
    instance._run_storage.add_run_telemetry(  # noqa: SLF001
        telemetry_data, tags=run_tags
    )


def _capture_metrics(
    instance: DagsterInstance,
    dagster_run: DagsterRun,
    container_metrics_enabled: bool,
    python_metrics_enabled: bool,
    shutdown_event: threading.Event,
    polling_interval: float = DEFAULT_RUN_METRICS_POLL_INTERVAL_SECONDS,
    logger: Optional[logging.Logger] = None,
) -> bool:
    check.inst_param(instance, "instance", DagsterInstance)
    check.inst_param(dagster_run, "dagster_run", DagsterRun)
    check.bool_param(container_metrics_enabled, "container_metrics_enabled")
    check.bool_param(python_metrics_enabled, "python_metrics_enabled")
    check.inst_param(shutdown_event, "shutdown_event", threading.Event)
    check.float_param(polling_interval, "polling_interval")
    check.opt_inst_param(logger, "logger", logging.Logger)

    if not (container_metrics_enabled or python_metrics_enabled):
        if logger:
            logger.warning("No metrics to capture, shutting down metrics capture thread")
        return False  # terminate the thread safely without interrupting the main thread

    run_tags = _metric_tags(dagster_run)

    if logger:
        logger.debug(f"Starting metrics capture thread with tags: {run_tags}")
        logger.debug(f"  [container_metrics_enabled={container_metrics_enabled}]")
        logger.debug(f"  [python_metrics_enabled={python_metrics_enabled}]")

    previous_cpu_usage_ms = None
    previous_measurement_timestamp = None
    while not shutdown_event.is_set():
        try:
            metrics = {}

            if container_metrics_enabled:
                container_metrics = _get_container_metrics(
                    previous_cpu_usage_ms=previous_cpu_usage_ms,
                    previous_measurement_timestamp=previous_measurement_timestamp,
                    logger=logger,
                )
                previous_cpu_usage_ms = container_metrics.get("container.cpu_usage_ms")
                previous_measurement_timestamp = container_metrics.get("measurement_timestamp")
                metrics.update(container_metrics)

            if python_metrics_enabled:
                python_metrics = _get_python_runtime_metrics()
                metrics.update(python_metrics)

            if len(metrics) > 0:
                _report_run_metrics(
                    instance,
                    dagster_run=dagster_run,
                    metrics=metrics,
                    run_tags=run_tags,
                )

        except:
            if logger:
                logger.warning(
                    "Exception caught during capture of metrics, will cease capturing",
                    exc_info=True,
                )
            return False  # terminate the thread safely without interrupting the main thread

        shutdown_event.wait(polling_interval)
    if logger:
        logger.debug("Shutting down metrics capture thread")
    return True


def start_run_metrics_thread(
    instance: DagsterInstance,
    dagster_run: DagsterRun,
    python_metrics_enabled: Optional[bool] = False,
    logger: Optional[logging.Logger] = None,
    polling_interval: float = DEFAULT_RUN_METRICS_POLL_INTERVAL_SECONDS,
) -> tuple[Optional[threading.Thread], Optional[threading.Event]]:
    check.inst_param(instance, "instance", DagsterInstance)
    check.inst_param(dagster_run, "dagster_run", DagsterRun)
    check.opt_inst_param(logger, "logger", logging.Logger)
    check.opt_bool_param(python_metrics_enabled, "python_metrics_enabled")
    check.float_param(polling_interval, "polling_interval")

    if not instance.run_storage.supports_run_telemetry():
        if logger:
            logger.debug("Run telemetry is not supported, skipping run metrics thread")
        return None, None

    container_metrics_enabled = _process_is_containerized()
    if not container_metrics_enabled and not python_metrics_enabled:
        if logger:
            logger.info(
                "No collectable metrics, skipping run metrics thread. "
                "This feature is not supported by some agents and requires execution in compatible linux "
                "containers based on cgroup v1 or v2."
            )
        return None, None

    if logger:
        logger.debug("Starting run metrics thread")

    instance.report_engine_event(
        f"Starting run metrics thread with container_metrics_enabled={container_metrics_enabled} and "
        f"python_metrics_enabled={python_metrics_enabled}",
        dagster_run=dagster_run,
        run_id=dagster_run.run_id,
    )

    shutdown_event = threading.Event()
    thread = threading.Thread(
        target=_capture_metrics,
        args=(
            instance,
            dagster_run,
            container_metrics_enabled,
            python_metrics_enabled,
            shutdown_event,
            polling_interval,
            logger,
        ),
        name="run-metrics",
        daemon=True,
    )
    thread.start()
    return thread, shutdown_event


def stop_run_metrics_thread(
    thread: threading.Thread,
    stop_event: threading.Event,
    timeout: Optional[int] = DEFAULT_RUN_METRICS_SHUTDOWN_SECONDS,
) -> bool:
    thread = check.not_none(thread)
    stop_event = check.not_none(stop_event)

    stop_event.set()
    if thread.is_alive():
        thread.join(timeout=timeout)

    return not thread.is_alive()



from dagster._serdes.config_class import (
    ConfigurableClass as ConfigurableClass,
    ConfigurableClassData as ConfigurableClassData,
    class_from_code_pointer as class_from_code_pointer,
)
from dagster._serdes.serdes import (
    EnumSerializer as EnumSerializer,
    NamedTupleSerializer as NamedTupleSerializer,
    SerializableNonScalarKeyMapping as SerializableNonScalarKeyMapping,
    WhitelistMap as WhitelistMap,
    deserialize_value as deserialize_value,
    deserialize_values as deserialize_values,
    get_storage_name as get_storage_name,
    pack_value as pack_value,
    serialize_value as serialize_value,
    unpack_value as unpack_value,
    whitelist_for_serdes as whitelist_for_serdes,
)
from dagster._serdes.utils import (
    create_snapshot_id as create_snapshot_id,
    serialize_pp as serialize_pp,
)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, AbstractSet, Any, Callable, Iterator, Optional

from typing_extensions import TypeAlias

import dagster._check as check
from dagster._config import ConfigType
from dagster._core.decorator_utils import get_function_params, validate_expected_params
from dagster._core.definitions.resource_requirement import (
    ResourceRequirement,
    TypeLoaderResourceRequirement,
)
from dagster._core.errors import DagsterInvalidDefinitionError

if TYPE_CHECKING:
    from dagster._core.execution.context.system import DagsterTypeLoaderContext


class DagsterTypeLoader(ABC):
    """Dagster type loaders are used to load unconnected inputs of the dagster type they are attached
    to.

    The recommended way to define a type loader is with the
    :py:func:`@dagster_type_loader <dagster_type_loader>` decorator.
    """

    @property
    @abstractmethod
    def schema_type(self) -> ConfigType:
        pass

    def construct_from_config_value(
        self, _context: "DagsterTypeLoaderContext", config_value: object
    ) -> object:
        """How to create a runtime value from config data."""
        return config_value

    def required_resource_keys(self) -> AbstractSet[str]:
        return frozenset()

    def get_resource_requirements(
        self, type_display_name: str
    ) -> Iterator[ResourceRequirement]:
        for resource_key in sorted(list(self.required_resource_keys())):
            yield TypeLoaderResourceRequirement(
                key=resource_key, type_display_name=type_display_name
            )


class DagsterTypeLoaderFromDecorator(DagsterTypeLoader):
    def __init__(self, config_type, func, required_resource_keys):
        self._config_type = check.inst_param(config_type, "config_type", ConfigType)
        self._func = check.callable_param(func, "func")
        self._required_resource_keys = check.opt_set_param(
            required_resource_keys, "required_resource_keys", of_type=str
        )

    @property
    def schema_type(self) -> ConfigType:
        return self._config_type

    def construct_from_config_value(
        self, context: "DagsterTypeLoaderContext", config_value: object
    ):
        return self._func(context, config_value)

    def required_resource_keys(self):
        return frozenset(self._required_resource_keys)


def _create_type_loader_for_decorator(
    config_type: ConfigType, func, required_resource_keys: Optional[AbstractSet[str]]
):
    return DagsterTypeLoaderFromDecorator(config_type, func, required_resource_keys)


DagsterTypeLoaderFn: TypeAlias = Callable[["DagsterTypeLoaderContext", Any], Any]


def dagster_type_loader(
    config_schema: object, required_resource_keys: Optional[AbstractSet[str]] = None
) -> Callable[[DagsterTypeLoaderFn], DagsterTypeLoaderFromDecorator]:
    """Create an dagster type loader that maps config data to a runtime value.

    The decorated function should take the execution context and parsed config value and return the
    appropriate runtime value.

    Args:
        config_schema (ConfigSchema): The schema for the config that's passed to the decorated
            function.

    Examples:
        .. code-block:: python

            @dagster_type_loader(Permissive())
            def load_dict(_context, value):
                return value
    """
    from dagster._config import resolve_to_config_type

    config_type = resolve_to_config_type(config_schema)
    assert isinstance(
        config_type, ConfigType
    ), f"{config_schema} could not be resolved to config type"
    EXPECTED_POSITIONALS = ["context", "*"]

    def wrapper(func: DagsterTypeLoaderFn) -> DagsterTypeLoaderFromDecorator:
        params = get_function_params(func)
        missing_positional = validate_expected_params(params, EXPECTED_POSITIONALS)
        if missing_positional:
            raise DagsterInvalidDefinitionError(
                f"@dagster_type_loader '{func.__name__}' decorated function does not have required"
                f" positional parameter '{missing_positional}'. @dagster_type_loader decorated"
                " functions should only have keyword arguments that match input names and a first"
                " positional parameter named 'context'."
            )

        return _create_type_loader_for_decorator(
            config_type, func, required_resource_keys
        )

    return wrapper



from collections.abc import Iterator, Mapping
from typing import Callable, NamedTuple, NoReturn, cast

from typing_extensions import TypeAlias

import dagster._check as check
from dagster._config import EvaluateValueResult, process_config
from dagster._core.definitions.asset_layer import AssetLayer
from dagster._core.definitions.dependency import GraphNode, Node, NodeHandle, OpNode
from dagster._core.definitions.graph_definition import GraphDefinition, SubselectedGraphDefinition
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.run_config import define_node_shape
from dagster._core.errors import (
    DagsterConfigMappingFunctionError,
    DagsterInvalidConfigError,
    user_code_error_boundary,
)
from dagster._core.system_config.objects import OpConfig
from dagster._utils.merger import merge_dicts

RawNodeConfig: TypeAlias = Mapping[str, object]


class OpConfigEntry(
    NamedTuple("_SolidConfigEntry", [("handle", NodeHandle), ("solid_config", OpConfig)])
):
    def __new__(cls, handle: NodeHandle, op_config: OpConfig):
        return super().__new__(
            cls,
            check.inst_param(handle, "handle", NodeHandle),
            check.inst_param(op_config, "solid_config", OpConfig),
        )


# This is a dummy handle used to simplify the code, corresponding to the root container (graph). It
# doesn't actually represent a node during execution.
_ROOT_HANDLE = NodeHandle("root", None)


class DescentStack(
    NamedTuple("_DescentStack", [("job_def", JobDefinition), ("handle", NodeHandle)])
):
    def __new__(cls, job_def: JobDefinition, handle: NodeHandle):
        return super().__new__(
            cls,
            job_def=check.inst_param(job_def, "job_def", JobDefinition),
            handle=check.inst_param(handle, "handle", NodeHandle),
        )

    @property
    def current_container(self) -> GraphDefinition:
        if self.handle == _ROOT_HANDLE:
            return self.job_def.graph
        else:
            assert isinstance(self.current_node, GraphNode)
            return self.current_node.definition

    @property
    def current_node(self) -> Node:
        assert self.handle is not None
        return self.job_def.get_node(self.handle)

    @property
    def current_handle_str(self) -> str:
        return str(check.not_none(self.handle))

    def descend(self, node: Node) -> "DescentStack":
        parent = self.handle if self.handle != _ROOT_HANDLE else None
        return self._replace(handle=NodeHandle(node.name, parent=parent))


def composite_descent(
    job_def: JobDefinition,
    ops_config: Mapping[str, RawNodeConfig],
    resource_defs: Mapping[str, ResourceDefinition],
) -> Mapping[str, OpConfig]:
    """This function is responsible for constructing the dictionary of OpConfig (indexed by handle)
    that will be passed into the ResolvedRunConfig. Critically this is the codepath that manages
    config mapping, where the runtime calls into user-defined config mapping functions to produce
    config for child nodes of graphs.

    Args:
        job_def (JobDefinition): JobDefinition
        ops_config (dict): Configuration for the ops in the pipeline. The "ops" entry
            of the run_config. Assumed to have already been validated.

    Returns:
        Dict[str, OpConfig]: A dictionary mapping string representations of NodeHandles to
            OpConfig objects. It includes an entry for ops at every level of the
            composite tree - i.e. not just leaf ops, but composite ops as well
    """
    check.inst_param(job_def, "job_def", JobDefinition)
    check.dict_param(ops_config, "ops_config")
    check.dict_param(resource_defs, "resource_defs", key_type=str, value_type=ResourceDefinition)

    # If top-level graph has config mapping, apply that config mapping before descending.
    if job_def.graph.has_config_mapping:
        ops_config = _apply_top_level_config_mapping(
            job_def,
            ops_config,
            resource_defs,
        )

    return {
        str(handle): op_config
        for handle, op_config in _composite_descent(
            parent_stack=DescentStack(job_def, _ROOT_HANDLE),
            ops_config_dict=ops_config,
            resource_defs=resource_defs,
            asset_layer=job_def.asset_layer,
        )
    }


def _composite_descent(
    parent_stack: DescentStack,
    ops_config_dict: Mapping[str, RawNodeConfig],
    resource_defs: Mapping[str, ResourceDefinition],
    asset_layer: AssetLayer,
) -> Iterator[OpConfigEntry]:
    """The core implementation of composite_descent. This yields a stream of OpConfigEntry. This is
    used by composite_descent to construct a dictionary.

    It descends over the entire node hierarchy, constructing an entry for every handle. If it
    encounters a graph instance with a config mapping, it will invoke that config mapping fn,
    producing the config that is necessary to configure the child nodes.

    This process unrolls recursively as you descend down the tree.
    """
    for node in parent_stack.current_container.nodes:
        current_stack = parent_stack.descend(node)
        current_handle = current_stack.handle

        current_op_config = ops_config_dict.get(node.name, {})

        # the base case
        if isinstance(node, OpNode):
            config_mapped_node_config = node.definition.apply_config_mapping(
                {"config": current_op_config.get("config")}
            )
            if not config_mapped_node_config.success:
                raise DagsterInvalidConfigError(
                    f"Error in config for {node.describe_node()}".format(node.name),
                    config_mapped_node_config.errors,
                    config_mapped_node_config,
                )

            complete_config_object = merge_dicts(
                current_op_config,
                config_mapped_node_config.value,  # type: ignore  # (unknown EVR type)
            )
            yield OpConfigEntry(current_handle, OpConfig.from_dict(complete_config_object))
            continue

        elif isinstance(node, GraphNode):
            yield OpConfigEntry(
                current_handle,
                OpConfig.from_dict(
                    {
                        "inputs": current_op_config.get("inputs"),
                        "outputs": current_op_config.get("outputs"),
                    }
                ),
            )

            # If there is a config mapping, invoke it and get the descendent nodes
            # config that way. Else just grabs the ops entry of the current config
            mapped_nodes_config = (
                _apply_config_mapping(
                    node,
                    current_stack,
                    current_op_config,
                    resource_defs,
                    asset_layer,
                )
                if node.definition.has_config_mapping
                else cast(Mapping[str, RawNodeConfig], current_op_config.get("ops", {}))
            )

            yield from _composite_descent(
                current_stack,
                mapped_nodes_config,
                resource_defs,
                asset_layer,
            )
        else:
            check.failed(f"Unexpected node type {type(node)}")


def _apply_top_level_config_mapping(
    job_def: JobDefinition,
    outer_config: Mapping[str, Mapping[str, object]],
    resource_defs: Mapping[str, ResourceDefinition],
) -> Mapping[str, RawNodeConfig]:
    graph_def = job_def.graph
    config_mapping = graph_def.config_mapping
    if config_mapping is None:
        return outer_config

    else:
        mapped_config_evr = graph_def.apply_config_mapping(outer_config)
        if not mapped_config_evr.success:
            raise DagsterInvalidConfigError(
                f"Error in config for graph {graph_def.name}",
                mapped_config_evr.errors,
                outer_config,
            )

        with user_code_error_boundary(
            DagsterConfigMappingFunctionError, _get_top_level_error_lambda(job_def)
        ):
            mapped_graph_config = config_mapping.resolve_from_validated_config(
                mapped_config_evr.value.get("config", {})  # type: ignore  # (possible none)
            )

        # Dynamically construct the type that the output of the config mapping function will
        # be evaluated against

        type_to_evaluate_against = define_node_shape(
            nodes=graph_def.nodes,
            ignored_nodes=None,
            dependency_structure=graph_def.dependency_structure,
            resource_defs=resource_defs,
            asset_layer=job_def.asset_layer,
            input_assets=graph_def.input_assets,
        )

        # process against that new type

        evr = process_config(type_to_evaluate_against, mapped_graph_config)

        if not evr.success:
            raise_top_level_config_error(job_def, mapped_graph_config, evr)

        return evr.value  # type: ignore  # (unknown evr type)


def _apply_config_mapping(
    graph_node: GraphNode,
    current_stack: DescentStack,
    current_node_config: RawNodeConfig,
    resource_defs: Mapping[str, ResourceDefinition],
    asset_layer: AssetLayer,
) -> Mapping[str, RawNodeConfig]:
    # the spec of the config mapping function is that it takes the dictionary at:
    # op_name:
    #    config: {dict_passed_to_user}

    # and it returns the dictionary rooted at ops
    # op_name:
    #    ops: {return_value_of_config_fn}

    # We must call the config mapping function and then validate it against
    # the child schema.

    # apply @configured config mapping to the composite's incoming config before we get to the
    # composite's own config mapping process
    graph_def = graph_node.definition
    config_mapped_node_config = graph_def.apply_config_mapping(current_node_config)
    if not config_mapped_node_config.success:
        raise DagsterInvalidConfigError(
            f"Error in config for graph {graph_node.name}",
            config_mapped_node_config.errors,
            config_mapped_node_config,
        )

    with user_code_error_boundary(
        DagsterConfigMappingFunctionError, _get_error_lambda(current_stack)
    ):
        config_mapping = check.not_none(graph_def.config_mapping)
        mapped_ops_config = config_mapping.resolve_from_validated_config(
            config_mapped_node_config.value.get("config", {})  # type: ignore  # (unknown EVR type)
        )

    # Dynamically construct the type that the output of the config mapping function will
    # be evaluated against

    # diff original graph and the subselected graph to find nodes to ignore so the system knows to
    # skip the validation then when config mapping generates values where the nodes are not selected
    ignored_nodes = (
        graph_def.get_top_level_omitted_nodes()
        if isinstance(graph_def, SubselectedGraphDefinition)
        else None
    )

    type_to_evaluate_against = define_node_shape(
        nodes=graph_def.nodes,
        ignored_nodes=ignored_nodes,
        dependency_structure=graph_def.dependency_structure,
        parent_handle=current_stack.handle,
        resource_defs=resource_defs,
        asset_layer=asset_layer,
        input_assets=graph_def.input_assets,
    )

    # process against that new type

    evr = process_config(type_to_evaluate_against, mapped_ops_config)

    if not evr.success:
        raise_composite_descent_config_error(current_stack, mapped_ops_config, evr)

    return evr.value  # type: ignore  # (unknown evr type)


def _get_error_lambda(current_stack: DescentStack) -> Callable[[], str]:
    return lambda: (
        "The config mapping function on {described_node} in {described_target} "
        "has thrown an unexpected error during its execution. The definition is "
        'instantiated at stack "{stack_str}".'
    ).format(
        described_node=current_stack.current_node.describe_node(),
        described_target=current_stack.job_def.describe_target(),
        stack_str=":".join(current_stack.handle.path),
    )


def _get_top_level_error_lambda(job_def: JobDefinition) -> Callable[[], str]:
    return lambda: (
        f"The config mapping function on top-level graph {job_def.graph.name} in job"
        f" {job_def.name} has thrown an unexpected error during its execution."
    )


def raise_top_level_config_error(
    job_def: JobDefinition, failed_config_value: object, evr: EvaluateValueResult
) -> NoReturn:
    message = (
        f"In job '{job_def.name}', top level graph '{job_def.graph.name}' has a "
        "configuration error."
    )

    raise DagsterInvalidConfigError(message, evr.errors, failed_config_value)


def raise_composite_descent_config_error(
    descent_stack: DescentStack, failed_config_value: object, evr: EvaluateValueResult
) -> NoReturn:
    check.inst_param(descent_stack, "descent_stack", DescentStack)
    check.inst_param(evr, "evr", EvaluateValueResult)

    node = descent_stack.current_node
    message = "In job {job_name} at stack {stack}: \n".format(
        job_name=descent_stack.job_def.name,
        stack=":".join(descent_stack.handle.path),
    )
    message += (
        f'Op "{node.name}" with definition "{node.definition.name}" has a '
        "configuration error. "
        "It has produced config a via its config_fn that fails to "
        "pass validation in the ops that it contains. "
        "This indicates an error in the config mapping function itself. It must "
        "produce correct config for its constituent ops in all cases. The correct "
        "resolution is to fix the mapping function. Details on the error (and the paths "
        'on this error are relative to config mapping function "root", not the entire document): '
    )

    raise DagsterInvalidConfigError(message, evr.errors, failed_config_value)



from collections.abc import Mapping
from typing import Any, Optional, Union

import dagster._check as check
from dagster._core.definitions import JobDefinition
from dagster._core.definitions.run_config import RunConfig, convert_config_input
from dagster._core.system_config.objects import ResolvedRunConfig


def validate_run_config(
    job_def: JobDefinition,
    run_config: Optional[Union[Mapping[str, Any], RunConfig]] = None,
) -> Mapping[str, Any]:
    """Function to validate a provided run config blob against a given job.

    If validation is successful, this function will return a dictionary representation of the
    validated config actually used during execution.

    Args:
        job_def (JobDefinition): The job definition to validate run
            config against
        run_config (Optional[Dict[str, Any]]): The run config to validate

    Returns:
        Dict[str, Any]: A dictionary representation of the validated config.
    """
    check.inst_param(job_def, "job_def", JobDefinition)
    run_config = check.opt_mapping_param(
        convert_config_input(run_config), "run_config", key_type=str
    )

    return ResolvedRunConfig.build(job_def, run_config).to_dict()



import itertools
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from functools import cached_property
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    AbstractSet,
    Annotated,
    Generic,
    Optional,
    TypeVar,
    Union,
)

import dagster._check as check
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_job import IMPLICIT_ASSET_JOB_NAME
from dagster._core.definitions.asset_key import EntityKey
from dagster._core.definitions.asset_spec import AssetExecutionType
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.base_asset_graph import (
    AssetCheckNode,
    AssetKey,
    BaseAssetGraph,
    BaseAssetNode,
)
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.metadata import ArbitraryMetadataMapping
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_mapping import PartitionMapping
from dagster._core.definitions.utils import DEFAULT_GROUP_NAME
from dagster._core.remote_representation.external import RemoteRepository
from dagster._core.remote_representation.handle import InstigatorHandle, RepositoryHandle
from dagster._core.workspace.workspace import CurrentWorkspace
from dagster._record import ImportFrom, record
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.cached_method import cached_method

if TYPE_CHECKING:
    from dagster._core.remote_representation.external_data import AssetCheckNodeSnap, AssetNodeSnap


@whitelist_for_serdes
@record
class RemoteAssetCheckNode:
    handle: RepositoryHandle
    asset_check: Annotated[
        "AssetCheckNodeSnap",
        ImportFrom("dagster._core.remote_representation.external_data"),
    ]
    execution_set_entity_keys: AbstractSet[EntityKey]


class RemoteAssetNode(BaseAssetNode, ABC):
    @abstractmethod
    def resolve_to_singular_repo_scoped_node(self) -> "RemoteRepositoryAssetNode": ...

    @property
    def execution_set_asset_keys(self) -> AbstractSet[AssetKey]:
        return {k for k in self.execution_set_entity_keys if isinstance(k, AssetKey)}

    @property
    def description(self) -> Optional[str]:
        return self.resolve_to_singular_repo_scoped_node().asset_node_snap.description

    @property
    def group_name(self) -> str:
        return (
            self.resolve_to_singular_repo_scoped_node().asset_node_snap.group_name
            or DEFAULT_GROUP_NAME
        )

    @property
    def metadata(self) -> ArbitraryMetadataMapping:
        return self.resolve_to_singular_repo_scoped_node().asset_node_snap.metadata

    @property
    def pools(self) -> Optional[set[str]]:
        return self.resolve_to_singular_repo_scoped_node().pools

    @property
    def tags(self) -> Mapping[str, str]:
        return self.resolve_to_singular_repo_scoped_node().asset_node_snap.tags or {}

    @property
    def owners(self) -> Sequence[str]:
        return self.resolve_to_singular_repo_scoped_node().asset_node_snap.owners or []

    @property
    def is_partitioned(self) -> bool:
        return self.resolve_to_singular_repo_scoped_node().asset_node_snap.partitions is not None

    @cached_property
    def partitions_def(self) -> Optional[PartitionsDefinition]:
        partitions_snap = self.resolve_to_singular_repo_scoped_node().asset_node_snap.partitions
        return partitions_snap.get_partitions_definition() if partitions_snap else None

    @property
    def freshness_policy(self) -> Optional[FreshnessPolicy]:
        # It is currently not possible to access the freshness policy for an observation definition
        # if a materialization definition also exists. This needs to be fixed.
        return self.resolve_to_singular_repo_scoped_node().asset_node_snap.freshness_policy

    @property
    def code_version(self) -> Optional[str]:
        # It is currently not possible to access the code version for an observation definition if a
        # materialization definition also exists. This needs to be fixed.
        return self.resolve_to_singular_repo_scoped_node().asset_node_snap.code_version

    @property
    def job_names(self) -> Sequence[str]:
        # It is currently not possible to access the job names for an observation definition if a
        # materialization definition also exists. This needs to be fixed.
        return (
            self.resolve_to_singular_repo_scoped_node().asset_node_snap.job_names
            if self.is_executable
            else []
        )


@whitelist_for_serdes
@record
class RemoteRepositoryAssetNode(RemoteAssetNode):
    """Asset nodes from a single RemoteRepository."""

    repository_handle: RepositoryHandle
    asset_node_snap: Annotated[
        "AssetNodeSnap",
        ImportFrom("dagster._core.remote_representation.external_data"),
    ]
    parent_keys: AbstractSet[AssetKey]
    child_keys: AbstractSet[AssetKey]
    check_keys: AbstractSet[AssetCheckKey]
    execution_set_entity_keys: AbstractSet[EntityKey]

    def __hash__(self):
        # we create sets of these objects in the context of asset graphs but don't want to
        # enforce that all recursively contained types are hashable so use object hash instead
        return object.__hash__(self)

    def resolve_to_singular_repo_scoped_node(self) -> "RemoteRepositoryAssetNode":
        return self

    @property
    def key(self) -> AssetKey:
        return self.asset_node_snap.asset_key

    @property
    def is_materializable(self) -> bool:
        return self.asset_node_snap.is_materializable

    @property
    def is_observable(self) -> bool:
        return self.asset_node_snap.is_observable

    @property
    def is_external(self) -> bool:
        return self.asset_node_snap.is_external

    @property
    def is_executable(self) -> bool:
        return self.asset_node_snap.is_executable

    @property
    def partition_mappings(self) -> Mapping[AssetKey, PartitionMapping]:
        return {
            dep.parent_asset_key: dep.partition_mapping
            for dep in self.asset_node_snap.parent_edges
            if dep.partition_mapping is not None
        }

    @property
    def backfill_policy(self) -> Optional[BackfillPolicy]:
        return self.asset_node_snap.backfill_policy

    @property
    def automation_condition(self) -> Optional[AutomationCondition]:
        return self.asset_node_snap.automation_condition

    @property
    def auto_materialize_policy(self) -> Optional[AutoMaterializePolicy]:
        return self.asset_node_snap.auto_materialize_policy

    @property
    def auto_observe_interval_minutes(self) -> Optional[float]:
        return self.asset_node_snap.auto_observe_interval_minutes

    @property
    def pools(self) -> Optional[set[str]]:
        return self.asset_node_snap.pools


@whitelist_for_serdes
@record
class RepositoryScopedAssetInfo:
    """RemoteRepositoryAssetNode paired with additional information from that repository.
    This split allows repository scoped asset graph to be constructed without depending on schedules/sensors
    as defining schedule/sensors needs an asset graph.
    """

    asset_node: RemoteRepositoryAssetNode
    targeting_sensor_names: Sequence[str]
    targeting_schedule_names: Sequence[str]

    @property
    def handle(self) -> RepositoryHandle:
        return self.asset_node.repository_handle


@whitelist_for_serdes
@record
class RemoteWorkspaceAssetNode(RemoteAssetNode):
    """Asset nodes constructed from a CurrentWorkspace, containing nodes from potentially several RemoteRepositories."""

    repo_scoped_asset_infos: Sequence[RepositoryScopedAssetInfo]

    def __hash__(self):
        # we create sets of these objects in the context of asset graphs but don't want to
        # enforce that all recursively contained types are hashable so use object hash instead
        return object.__hash__(self)

    ##### COMMON ASSET NODE INTERFACE
    @cached_property
    def key(self) -> AssetKey:
        return self.repo_scoped_asset_infos[0].asset_node.asset_node_snap.asset_key

    @property
    def parent_keys(self) -> AbstractSet[AssetKey]:
        # combine deps from all nodes
        keys = set()
        for info in self.repo_scoped_asset_infos:
            keys.update(info.asset_node.parent_keys)
        return keys

    @property
    def child_keys(self) -> AbstractSet[AssetKey]:
        # combine deps from all nodes
        keys = set()
        for info in self.repo_scoped_asset_infos:
            keys.update(info.asset_node.child_keys)
        return keys

    @property
    def check_keys(self) -> AbstractSet[AssetCheckKey]:
        # combine check keys from all nodes
        keys = set()
        for info in self.repo_scoped_asset_infos:
            keys.update(info.asset_node.check_keys)
        return keys

    @property
    def execution_set_entity_keys(
        self,
    ) -> AbstractSet[Union[AssetKey, AssetCheckKey]]:
        return self.resolve_to_singular_repo_scoped_node().execution_set_entity_keys

    @cached_property
    def is_materializable(self) -> bool:
        return any(info.asset_node.is_materializable for info in self.repo_scoped_asset_infos)

    @cached_property
    def is_observable(self) -> bool:
        return any(info.asset_node.is_observable for info in self.repo_scoped_asset_infos)

    @cached_property
    def is_external(self) -> bool:
        return all(info.asset_node.is_external for info in self.repo_scoped_asset_infos)

    @cached_property
    def is_executable(self) -> bool:
        return any(node.asset_node.is_executable for node in self.repo_scoped_asset_infos)

    @cached_property
    def pools(self) -> Optional[set[str]]:
        pools = set()
        for info in self.repo_scoped_asset_infos:
            pools.update(info.asset_node.pools or set())
        return pools

    @property
    def partition_mappings(self) -> Mapping[AssetKey, PartitionMapping]:
        if self.is_materializable:
            return {
                dep.parent_asset_key: dep.partition_mapping
                for dep in self._materializable_node_snap.parent_edges
                if dep.partition_mapping is not None
            }
        else:
            return {}

    @property
    def auto_materialize_policy(self) -> Optional[AutoMaterializePolicy]:
        return (
            self._materializable_node_snap.auto_materialize_policy
            if self.is_materializable
            else None
        )

    @property
    def automation_condition(self) -> Optional[AutomationCondition]:
        if self.is_materializable:
            return self._materializable_node_snap.automation_condition
        elif self.is_observable:
            return self._observable_node_snap.automation_condition
        else:
            return None

    @property
    def auto_observe_interval_minutes(self) -> Optional[float]:
        return (
            self._observable_node_snap.auto_observe_interval_minutes if self.is_observable else None
        )

    @property
    def backfill_policy(self) -> Optional[BackfillPolicy]:
        return self._materializable_node_snap.backfill_policy if self.is_materializable else None

    ##### REMOTE-SPECIFIC INTERFACE
    @cached_method
    def resolve_to_singular_repo_scoped_node(self) -> "RemoteRepositoryAssetNode":
        # Return a materialization node if it exists, otherwise return an observable node if it
        # exists, otherwise return any node. This exists to preserve implicit behavior, where the
        # materialization node was previously preferred over the observable node. This is a
        # temporary measure until we can appropriately scope the accessors that could apply to
        # either a materialization or observation node.
        # This property supports existing behavior but it should be phased out, because it relies on
        # materialization nodes shadowing observation nodes that would otherwise be exposed.
        return next(
            itertools.chain(
                (
                    info.asset_node
                    for info in self.repo_scoped_asset_infos
                    if info.asset_node.is_materializable
                ),
                (
                    info.asset_node
                    for info in self.repo_scoped_asset_infos
                    if info.asset_node.is_observable
                ),
                (info.asset_node for info in self.repo_scoped_asset_infos),
            )
        )

    def get_targeting_schedule_handles(
        self,
    ) -> Sequence[InstigatorHandle]:
        selectors = []
        for node in self.repo_scoped_asset_infos:
            for schedule_name in node.targeting_schedule_names:
                selectors.append(
                    InstigatorHandle(
                        repository_handle=node.handle,
                        instigator_name=schedule_name,
                    )
                )

        return selectors

    def get_targeting_sensor_handles(
        self,
    ) -> Sequence[InstigatorHandle]:
        selectors = []
        for node in self.repo_scoped_asset_infos:
            for sensor_name in node.targeting_sensor_names:
                selectors.append(
                    InstigatorHandle(
                        repository_handle=node.handle,
                        instigator_name=sensor_name,
                    )
                )
        return selectors

    ##### HELPERS

    @cached_property
    def _materializable_node_snap(self) -> "AssetNodeSnap":
        try:
            return next(
                info.asset_node.asset_node_snap
                for info in self.repo_scoped_asset_infos
                if info.asset_node.is_materializable
            )
        except StopIteration:
            check.failed("No materializable node found")

    @cached_property
    def _observable_node_snap(self) -> "AssetNodeSnap":
        try:
            return next(
                info.asset_node.asset_node_snap
                for info in self.repo_scoped_asset_infos
                if info.asset_node.is_observable
            )
        except StopIteration:
            check.failed("No observable node found")


TRemoteAssetNode = TypeVar("TRemoteAssetNode", bound=RemoteAssetNode)


class RemoteAssetGraph(BaseAssetGraph[TRemoteAssetNode], ABC, Generic[TRemoteAssetNode]):
    @property
    @abstractmethod
    def remote_asset_nodes_by_key(self) -> Mapping[AssetKey, TRemoteAssetNode]: ...

    @property
    @abstractmethod
    def remote_asset_check_nodes_by_key(
        self,
    ) -> Mapping[AssetCheckKey, RemoteAssetCheckNode]: ...

    ##### COMMON ASSET GRAPH INTERFACE
    @cached_property
    def _asset_check_nodes_by_key(self) -> Mapping[AssetCheckKey, AssetCheckNode]:
        return {
            k: AssetCheckNode(
                k,
                v.asset_check.additional_asset_keys,
                v.asset_check.blocking,
                v.asset_check.automation_condition,
            )
            for k, v in self.remote_asset_check_nodes_by_key.items()
        }

    def get_execution_set_asset_and_check_keys(
        self, entity_key: EntityKey
    ) -> AbstractSet[EntityKey]:
        if isinstance(entity_key, AssetKey):
            return self.get(entity_key).execution_set_entity_keys
        else:  # AssetCheckKey
            return self.remote_asset_check_nodes_by_key[entity_key].execution_set_entity_keys

    ##### REMOTE-SPECIFIC METHODS

    @property
    def asset_checks(self) -> Sequence["AssetCheckNodeSnap"]:
        return [node.asset_check for node in self.remote_asset_check_nodes_by_key.values()]

    @cached_property
    def _asset_check_nodes_by_asset_key(self) -> Mapping[AssetKey, Sequence[RemoteAssetCheckNode]]:
        by_asset_key = {}
        for node in self.remote_asset_check_nodes_by_key.values():
            by_asset_key.setdefault(node.asset_check.asset_key, []).append(node)
        return by_asset_key

    def get_checks_for_asset(self, asset_key: AssetKey) -> Sequence[RemoteAssetCheckNode]:
        return self._asset_check_nodes_by_asset_key.get(asset_key, [])

    def get_check_keys_for_assets(
        self, asset_keys: AbstractSet[AssetKey]
    ) -> AbstractSet[AssetCheckKey]:
        return (
            set().union(
                *(
                    {check.asset_check.key for check in self.get_checks_for_asset(asset_key)}
                    for asset_key in asset_keys
                )
            )
            if asset_keys
            else set()
        )

    @cached_property
    def asset_check_keys(self) -> AbstractSet[AssetCheckKey]:
        return set(self.remote_asset_check_nodes_by_key.keys())

    def asset_keys_for_job(self, job_name: str) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes if job_name in node.job_names}

    @cached_property
    def all_job_names(self) -> AbstractSet[str]:
        return {job_name for node in self.asset_nodes for job_name in node.job_names}

    def get_materialization_job_names(self, asset_key: AssetKey) -> Sequence[str]:
        """Returns the names of jobs that materialize this asset."""
        # This is a poorly named method because it will expose observation job names for assets with
        # a defined observation but no materialization.
        return self.get(asset_key).job_names

    def get_materialization_asset_keys_for_job(self, job_name: str) -> Sequence[AssetKey]:
        """Returns asset keys that are targeted for materialization in the given job."""
        return [
            k
            for k in self.materializable_asset_keys
            if job_name in self.get_materialization_job_names(k)
        ]

    def get_implicit_job_name_for_assets(
        self,
        asset_keys: Iterable[AssetKey],
        remote_repo: Optional[RemoteRepository],
    ) -> Optional[str]:
        """Returns the name of the asset base job that contains all the given assets, or None if there is no such
        job.

        Note: all asset_keys should be in the same repository.
        """
        return IMPLICIT_ASSET_JOB_NAME


@record
class RemoteRepositoryAssetGraph(RemoteAssetGraph[RemoteRepositoryAssetNode]):
    remote_asset_nodes_by_key: Mapping[AssetKey, RemoteRepositoryAssetNode]
    remote_asset_check_nodes_by_key: Mapping[AssetCheckKey, RemoteAssetCheckNode]

    @property
    def _asset_nodes_by_key(self) -> Mapping[AssetKey, RemoteRepositoryAssetNode]:
        return self.remote_asset_nodes_by_key

    @classmethod
    def build(cls, repo: RemoteRepository):
        # First pass, we need to:

        # * Build the dependency graph of asset keys.
        upstream: dict[AssetKey, set[AssetKey]] = defaultdict(set)
        downstream: dict[AssetKey, set[AssetKey]] = defaultdict(set)

        # * Build an index of execution sets by key. An execution set is a set of assets and checks
        # that must be executed together. AssetNodeSnaps and AssetCheckNodeSnaps already have an
        # optional execution_set_identifier set. A null execution_set_identifier indicates that the
        # node or check can be executed independently.
        execution_sets_by_id: dict[str, set[EntityKey]] = defaultdict(set)

        # * Map checks to their corresponding asset keys
        check_keys_by_asset_key: dict[AssetKey, set[AssetCheckKey]] = defaultdict(set)
        for asset_snap in repo.get_asset_node_snaps():
            id = asset_snap.execution_set_identifier
            key = asset_snap.asset_key
            if id is not None:
                execution_sets_by_id[id].add(key)

            for dep in asset_snap.parent_edges:
                upstream[asset_snap.asset_key].add(dep.parent_asset_key)
                downstream[dep.parent_asset_key].add(asset_snap.asset_key)

        for check_snap in repo.get_asset_check_node_snaps():
            id = check_snap.execution_set_identifier
            key = check_snap.key
            if id is not None:
                execution_sets_by_id[id].add(key)

            check_keys_by_asset_key[check_snap.asset_key].add(check_snap.key)

        # Second Pass - build the final nodes
        assets_by_key: dict[AssetKey, RemoteRepositoryAssetNode] = {}
        asset_checks_by_key: dict[AssetCheckKey, RemoteAssetCheckNode] = {}

        for asset_snap in repo.get_asset_node_snaps():
            id = asset_snap.execution_set_identifier
            key = asset_snap.asset_key

            assets_by_key[key] = RemoteRepositoryAssetNode(
                repository_handle=repo.handle,
                asset_node_snap=asset_snap,
                execution_set_entity_keys=execution_sets_by_id[id] if id is not None else {key},
                parent_keys=upstream[key],
                child_keys=downstream[key],
                check_keys=check_keys_by_asset_key[key],
            )

        for check_snap in repo.get_asset_check_node_snaps():
            id = check_snap.execution_set_identifier
            key = check_snap.key

            asset_checks_by_key[key] = RemoteAssetCheckNode(
                handle=repo.handle,
                asset_check=check_snap,
                execution_set_entity_keys=execution_sets_by_id[id] if id is not None else {key},
            )

        return cls(
            remote_asset_nodes_by_key=assets_by_key,
            remote_asset_check_nodes_by_key=asset_checks_by_key,
        )

    @classmethod
    def empty(cls):
        return cls(
            remote_asset_nodes_by_key={},
            remote_asset_check_nodes_by_key={},
        )


class RemoteWorkspaceAssetGraph(RemoteAssetGraph[RemoteWorkspaceAssetNode]):
    def __init__(
        self,
        remote_asset_nodes_by_key: Mapping[AssetKey, RemoteWorkspaceAssetNode],
        remote_asset_check_nodes_by_key: Mapping[AssetCheckKey, RemoteAssetCheckNode],
    ):
        self._remote_asset_nodes_by_key = remote_asset_nodes_by_key
        self._remote_asset_check_nodes_by_key = remote_asset_check_nodes_by_key

    @property
    def remote_asset_nodes_by_key(self) -> Mapping[AssetKey, RemoteWorkspaceAssetNode]:
        return self._remote_asset_nodes_by_key

    @property
    def remote_asset_check_nodes_by_key(
        self,
    ) -> Mapping[AssetCheckKey, RemoteAssetCheckNode]:
        return self._remote_asset_check_nodes_by_key

    @property
    def _asset_nodes_by_key(self) -> Mapping[AssetKey, RemoteWorkspaceAssetNode]:
        return self.remote_asset_nodes_by_key

    @property
    def asset_node_snaps_by_key(self) -> Mapping[AssetKey, "AssetNodeSnap"]:
        # This exists to support existing callsites but it should be removed ASAP, since it exposes
        # `AssetNodeSnap` instances directly. All sites using this should use RemoteAssetNode
        # instead.
        return {
            k: node.resolve_to_singular_repo_scoped_node().asset_node_snap
            for k, node in self._asset_nodes_by_key.items()
        }

    @cached_property
    def repository_handles_by_key(self) -> Mapping[EntityKey, RepositoryHandle]:
        return {
            **{
                k: node.resolve_to_singular_repo_scoped_node().repository_handle
                for k, node in self._asset_nodes_by_key.items()
            },
            **{k: v.handle for k, v in self.remote_asset_check_nodes_by_key.items()},
        }

    def get_repository_handle(self, key: EntityKey) -> RepositoryHandle:
        if isinstance(key, AssetKey):
            return self.get(key).resolve_to_singular_repo_scoped_node().repository_handle
        else:
            return self.remote_asset_check_nodes_by_key[key].handle

    def split_entity_keys_by_repository(
        self, keys: AbstractSet[EntityKey]
    ) -> Sequence[AbstractSet[EntityKey]]:
        keys_by_repo = defaultdict(set)
        for key in keys:
            repo_handle = self.get_repository_handle(key)
            keys_by_repo[(repo_handle.location_name, repo_handle.repository_name)].add(key)
        return list(keys_by_repo.values())

    @classmethod
    def build(cls, workspace: CurrentWorkspace):
        # Combine repository scoped asset graphs with additional context to form the global graph

        code_locations = sorted(
            (
                location_entry.code_location
                for location_entry in workspace.code_location_entries.values()
                if location_entry.code_location
            ),
            key=lambda code_location: code_location.name,
        )
        repos = (
            repo
            for code_location in code_locations
            for repo in sorted(
                code_location.get_repositories().values(), key=lambda repo: repo.name
            )
        )

        asset_infos_by_key: dict[AssetKey, list[RepositoryScopedAssetInfo]] = defaultdict(list)
        asset_checks_by_key: dict[AssetCheckKey, RemoteAssetCheckNode] = {}
        for repo in repos:
            for key, asset_node in repo.asset_graph.remote_asset_nodes_by_key.items():
                asset_infos_by_key[key].append(
                    RepositoryScopedAssetInfo(
                        asset_node=asset_node,
                        targeting_sensor_names=sorted(
                            s.name for s in repo.get_sensors_targeting(asset_node.key)
                        ),
                        targeting_schedule_names=sorted(
                            s.name for s in repo.get_schedules_targeting(asset_node.key)
                        ),
                    )
                )
            # NOTE: matches previous behavior of completely ignoring asset check collisions
            asset_checks_by_key.update(repo.asset_graph.remote_asset_check_nodes_by_key)

        asset_nodes_by_key = {}
        nodes_with_multiple = []
        for key, asset_infos in asset_infos_by_key.items():
            node = RemoteWorkspaceAssetNode(
                repo_scoped_asset_infos=asset_infos,
            )
            asset_nodes_by_key[key] = node
            if len(asset_infos) > 1:
                nodes_with_multiple.append(node)

        _warn_on_duplicate_nodes(nodes_with_multiple)

        return cls(
            remote_asset_nodes_by_key=asset_nodes_by_key,
            remote_asset_check_nodes_by_key=asset_checks_by_key,
        )


def _warn_on_duplicate_nodes(
    nodes_with_multiple: Sequence[RemoteWorkspaceAssetNode],
) -> None:
    # Split the nodes into materializable, observable, and unexecutable nodes. Observable and
    # unexecutable `AssetNodeSnap` represent both source and external assets-- the
    # "External" in "AssetNodeSnap" is unrelated to the "external" in "external asset", this
    # is just an unfortunate naming collision. `AssetNodeSnap` will be renamed eventually.
    materializable_duplicates: Mapping[AssetKey, Sequence[str]] = {}
    observable_duplicates: Mapping[AssetKey, Sequence[str]] = {}
    for node in nodes_with_multiple:
        check.invariant(
            len(node.repo_scoped_asset_infos) > 1,
            "only perform check on nodes with multiple defs",
        )
        observable_locations = []
        materializable_locations = []
        for info in node.repo_scoped_asset_infos:
            snap = info.asset_node.asset_node_snap
            location = info.asset_node.repository_handle.location_name
            if snap.is_source and snap.is_observable:
                observable_locations.append(location)
            elif snap.is_materializable:
                materializable_locations.append(location)
        if len(observable_locations) > 1:
            observable_duplicates[node.key] = observable_locations
        if len(materializable_locations) > 1:
            materializable_duplicates[node.key] = materializable_locations

    # It is possible for multiple nodes to exist that share the same key. This is invalid if
    # more than one node is materializable or if more than one node is observable. It is valid
    # if there is at most one materializable node and at most one observable node, with all
    # other nodes unexecutable.
    _warn_on_duplicates_within_subset(materializable_duplicates, AssetExecutionType.MATERIALIZATION)
    _warn_on_duplicates_within_subset(observable_duplicates, AssetExecutionType.OBSERVATION)


def _warn_on_duplicates_within_subset(
    duplicates: Mapping[AssetKey, Sequence[str]],
    execution_type: AssetExecutionType,
) -> None:
    if duplicates:
        duplicate_lines = []
        for asset_key, loc_names in duplicates.items():
            duplicate_lines.append(f"  {asset_key.to_string()}: {loc_names}")
        duplicate_str = "\n".join(duplicate_lines)

        warnings.warn(
            f"Found {execution_type.value} nodes for some asset keys in multiple code locations."
            f" Only one {execution_type.value} node is allowed per asset key. Duplicates:\n {duplicate_str}"
        )



import inspect
from collections.abc import Iterable, Iterator, Mapping
from typing import Any, Optional, Union, cast

from typing_extensions import TypedDict

import dagster._check as check
from dagster._core.definitions import (
    AssetCheckEvaluation,
    AssetCheckSeverity,
    AssetKey,
    AssetMaterialization,
    AssetObservation,
    ExpectationResult,
    Output,
    OutputDefinition,
    TypeCheck,
)
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_spec import AssetExecutionType
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.data_version import (
    CODE_VERSION_TAG,
    DATA_VERSION_IS_USER_PROVIDED_TAG,
    DATA_VERSION_TAG,
    DEFAULT_DATA_VERSION,
    NULL_EVENT_POINTER,
    DataVersion,
    compute_logical_data_version,
    get_input_data_version_tag,
    get_input_event_pointer_tag,
)
from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction
from dagster._core.definitions.events import DynamicOutput
from dagster._core.definitions.metadata import MetadataValue, normalize_metadata
from dagster._core.definitions.multi_dimensional_partitions import (
    MultiPartitionKey,
    get_tags_from_multi_partition_key,
)
from dagster._core.definitions.result import AssetResult
from dagster._core.definitions.source_asset import SYSTEM_METADATA_KEY_SOURCE_ASSET_OBSERVATION
from dagster._core.errors import (
    DagsterAssetCheckFailedError,
    DagsterExecutionHandleOutputError,
    DagsterInvariantViolationError,
    DagsterStepOutputNotFoundError,
    DagsterTypeCheckDidNotPass,
    DagsterTypeCheckError,
    user_code_error_boundary,
)
from dagster._core.events import DagsterEvent, DagsterEventBatchMetadata, generate_event_batch_id
from dagster._core.execution.context.compute import enter_execution_context
from dagster._core.execution.context.output import OutputContext
from dagster._core.execution.context.system import StepExecutionContext, TypeCheckContext
from dagster._core.execution.plan.compute import OpOutputUnion, execute_core_compute
from dagster._core.execution.plan.compute_generator import create_op_compute_wrapper
from dagster._core.execution.plan.inputs import StepInputData
from dagster._core.execution.plan.objects import StepSuccessData, TypeCheckData
from dagster._core.execution.plan.outputs import StepOutputData, StepOutputHandle
from dagster._core.execution.plan.utils import op_execution_error_boundary
from dagster._core.storage.tags import BACKFILL_ID_TAG
from dagster._core.types.dagster_type import DagsterType
from dagster._utils import iterate_with_context
from dagster._utils.timing import time_execution_scope
from dagster._utils.warnings import beta_warning, disable_dagster_warnings


class AssetResultOutput(Output):
    """This is a marker subclass that represents an Output that was produced from an AssetResult."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def _process_asset_results_to_events(
    step_context: StepExecutionContext,
    user_event_sequence: Iterator[OpOutputUnion],
) -> Iterator[OpOutputUnion]:
    """Handle converting MaterializeResult (& AssetCheckResult soon) to their appropriate events.

    MaterializeResults get converted to an Output event, which is later use to derive an AssetMaterialization.

    AssetCheckResult get converted to two events:
     - An Output, which allows downstream steps to depend on it
     - An AssetCheckEvaluation, which combines the check result with information from the context
         to create a full picture of the asset check's evaluation.
    """
    for user_event in user_event_sequence:
        yield from _process_user_event(step_context, user_event)


def _process_user_event(
    step_context: StepExecutionContext, user_event: OpOutputUnion
) -> Iterator[OpOutputUnion]:
    if isinstance(user_event, AssetResult):
        assets_def = _get_assets_def_for_step(step_context, user_event)
        asset_key = _resolve_asset_result_asset_key(user_event, assets_def)
        output_name = assets_def.get_output_name_for_asset_key(asset_key)

        for check_result in user_event.check_results or []:
            yield from _process_user_event(step_context, check_result)

        with disable_dagster_warnings():
            yield AssetResultOutput(
                value=None,
                output_name=output_name,
                metadata=user_event.metadata,
                data_version=user_event.data_version,
                tags=user_event.tags,
            )
    elif isinstance(user_event, AssetCheckResult):
        asset_check_evaluation = user_event.to_asset_check_evaluation(step_context)
        spec = check.not_none(
            step_context.job_def.asset_layer.asset_graph.get_check_spec(
                asset_check_evaluation.asset_check_key
            ),
            "If we were able to create an AssetCheckEvaluation from the AssetCheckResult, then"
            " there should be a spec for the check",
        )

        output_name = step_context.job_def.asset_layer.get_output_name_for_asset_check(
            asset_check_evaluation.asset_check_key
        )
        output = Output(value=None, output_name=output_name)

        yield asset_check_evaluation

        if (
            not asset_check_evaluation.passed
            and asset_check_evaluation.severity == AssetCheckSeverity.ERROR
            and spec.blocking
        ):
            raise DagsterAssetCheckFailedError(
                f"Blocking check '{spec.name}' for asset '{spec.asset_key.to_user_string()}' failed with"
                " ERROR severity."
            )
        yield output
    else:
        yield user_event


def _get_assets_def_for_step(
    step_context: StepExecutionContext, user_event: OpOutputUnion
) -> AssetsDefinition:
    assets_def = step_context.job_def.asset_layer.assets_def_for_node(step_context.node_handle)
    if not assets_def:
        raise DagsterInvariantViolationError(
            f"{user_event.__class__.__name__} is only valid within asset computations, no backing"
            " AssetsDefinition found."
        )
    return assets_def


def _resolve_asset_result_asset_key(
    asset_result: AssetResult, assets_def: AssetsDefinition
) -> AssetKey:
    if asset_result.asset_key:
        return asset_result.asset_key
    else:
        if len(assets_def.keys) != 1:
            raise DagsterInvariantViolationError(
                f"{asset_result.__class__.__name__} did not include asset_key and it can not be inferred."
                f" Specify which asset_key, options are: {assets_def.keys}."
            )
        return assets_def.key


def _step_output_error_checked_user_event_sequence(
    step_context: StepExecutionContext, user_event_sequence: Iterator[OpOutputUnion]
) -> Iterator[OpOutputUnion]:
    """Process the event sequence to check for invariant violations in the event
    sequence related to Output events emitted from the compute_fn.

    This consumes and emits an event sequence.
    """
    check.inst_param(step_context, "step_context", StepExecutionContext)
    check.iterator_param(user_event_sequence, "user_event_sequence")

    step = step_context.step
    op_label = step_context.describe_op()
    output_names = list([output_def.name for output_def in step.step_outputs])
    selected_output_names = step_context.selected_output_names

    for user_event in user_event_sequence:
        if not isinstance(user_event, (Output, DynamicOutput)):
            yield user_event
            continue

        # do additional processing on Outputs
        output = user_event
        if not step.has_step_output(cast(str, output.output_name)):
            raise DagsterInvariantViolationError(
                f'Core compute for {op_label} returned an output "{output.output_name}" that does '
                f"not exist. The available outputs are {output_names}"
            )

        if output.output_name not in selected_output_names:
            raise DagsterInvariantViolationError(
                f'Core compute for {op_label} returned an output "{output.output_name}" that is '
                f"not selected. The selected outputs are {selected_output_names}"
            )

        step_output = step.step_output_named(cast(str, output.output_name))
        output_def = step_context.job_def.get_node(step_output.node_handle).output_def_named(
            step_output.name
        )

        if isinstance(output, Output):
            if step_context.has_seen_output(output.output_name):
                raise DagsterInvariantViolationError(
                    f'Compute for {op_label} returned an output "{output.output_name}" multiple '
                    "times"
                )

            if output_def.is_dynamic:
                raise DagsterInvariantViolationError(
                    f'Compute for {op_label} for output "{output.output_name}" defined as dynamic '
                    "must yield DynamicOutput, got Output."
                )

            # For any output associated with an asset, make sure that none of its dependent assets
            # have already been yielded. If this condition (outputs yielded in topological order) is
            # not satisfied, automatic data version computation can yield wrong results.
            #
            # We look for dependent keys that have already been yielded rather than dependency keys
            # that have not yet been yielded. This is because we don't always know which
            # dependencies will actually be computed within the step. If A depends on B, it is
            # possible that a cached version of B will be used and B will never be yielded. In
            # contrast, if both A and B are yielded, A should never precede B.
            asset_layer = step_context.job_def.asset_layer
            node_handle = step_context.node_handle
            asset_key = asset_layer.asset_key_for_output(node_handle, output_def.name)
            if asset_key is not None and asset_key in asset_layer.asset_keys_for_node(node_handle):
                asset_node = asset_layer.get(asset_key)
                assets_def = asset_node.assets_def
                all_dependent_keys = asset_node.child_keys
                step_local_asset_keys = step_context.get_output_asset_keys()
                step_local_dependent_keys = all_dependent_keys & step_local_asset_keys
                for dependent_key in step_local_dependent_keys:
                    output_name = assets_def.get_output_name_for_asset_key(dependent_key)
                    # Need to skip self-dependent assets (possible with partitions)
                    self_dep = dependent_key in asset_node.parent_keys
                    if not self_dep and step_context.has_seen_output(output_name):
                        raise DagsterInvariantViolationError(
                            f'Asset "{dependent_key.to_user_string()}" was yielded before its'
                            f' dependency "{asset_key.to_user_string()}".Multiassets'
                            " yielding multiple asset outputs must yield them in topological"
                            " order."
                        )

            step_context.observe_output(output.output_name)

            metadata = step_context.get_output_metadata(output.output_name)
            with disable_dagster_warnings():
                output = output.with_metadata(
                    metadata={**output.metadata, **normalize_metadata(metadata or {})}
                )
        else:
            if not output_def.is_dynamic:
                raise DagsterInvariantViolationError(
                    f"Compute for {op_label} yielded a DynamicOutput, but did not use "
                    "DynamicOutputDefinition."
                )
            if step_context.has_seen_output(output.output_name, output.mapping_key):
                raise DagsterInvariantViolationError(
                    f"Compute for {op_label} yielded a DynamicOutput with mapping_key "
                    f'"{output.mapping_key}" multiple times.'
                )
            step_context.observe_output(output.output_name, output.mapping_key)
            metadata = step_context.get_output_metadata(
                output.output_name, mapping_key=output.mapping_key
            )
            output = DynamicOutput(
                value=output.value,
                output_name=output.output_name,
                metadata={**output.metadata, **normalize_metadata(metadata or {})},
                mapping_key=output.mapping_key,
            )

        yield output

    for step_output in step.step_outputs:
        step_output_def = step_context.op_def.output_def_named(step_output.name)
        if not step_context.has_seen_output(step_output_def.name) and not step_output_def.optional:
            asset_layer = step_context.job_def.asset_layer
            asset_key = asset_layer.asset_key_for_output(
                step_context.node_handle, step_output_def.name
            )
            # We require explicitly returned/yielded for asset observations
            is_observable_asset = asset_key is not None and asset_layer.get(asset_key).is_observable

            if step_output_def.dagster_type.is_nothing and not is_observable_asset:
                if step_output.name in selected_output_names:
                    step_context.log.info(
                        f'Emitting implicit Nothing for output "{step_output_def.name}" on {op_label}'
                    )
                    yield Output(output_name=step_output_def.name, value=None)
            elif not step_output_def.is_dynamic:
                raise DagsterStepOutputNotFoundError(
                    f"Core compute for {op_label} did not return an output for non-optional "
                    f'output "{step_output_def.name}"',
                    step_key=step.key,
                    output_name=step_output_def.name,
                )


def do_type_check(context: TypeCheckContext, dagster_type: DagsterType, value: Any) -> TypeCheck:
    type_check = dagster_type.type_check(context, value)
    if not isinstance(type_check, TypeCheck):
        return TypeCheck(
            success=False,
            description=(
                f"Type checks must return TypeCheck. Type check for type {dagster_type.display_name} returned "
                f"value of type {type(type_check)} when checking runtime value of type {type(value)}."
            ),
        )
    return type_check


def _create_step_input_event(
    step_context: StepExecutionContext, input_name: str, type_check: TypeCheck, success: bool
) -> DagsterEvent:
    return DagsterEvent.step_input_event(
        step_context,
        StepInputData(
            input_name=input_name,
            type_check_data=TypeCheckData(
                success=success,
                label=input_name,
                description=type_check.description if type_check else None,
                metadata=type_check.metadata if type_check else {},
            ),
        ),
    )


def _type_checked_event_sequence_for_input(
    step_context: StepExecutionContext,
    input_name: str,
    input_value: Any,
) -> Iterator[DagsterEvent]:
    check.inst_param(step_context, "step_context", StepExecutionContext)
    check.str_param(input_name, "input_name")

    step_input = step_context.step.step_input_named(input_name)
    input_def = step_context.op_def.input_def_named(step_input.name)

    check.invariant(
        input_def.name == input_name,
        f"InputDefinition name does not match, expected {input_name} got {input_def.name}",
    )

    dagster_type = input_def.dagster_type
    type_check_context = step_context.for_type(dagster_type)
    input_type = type(input_value)
    op_label = step_context.describe_op()

    with user_code_error_boundary(
        DagsterTypeCheckError,
        lambda: (
            f'Error occurred while type-checking input "{input_name}" of {op_label}, with Python'
            f" type {input_type} and Dagster type {dagster_type.display_name}"
        ),
        log_manager=type_check_context.log,
    ):
        type_check = do_type_check(type_check_context, dagster_type, input_value)

    yield _create_step_input_event(
        step_context, input_name, type_check=type_check, success=type_check.success
    )

    if not type_check.success:
        raise DagsterTypeCheckDidNotPass(
            description=(
                f'Type check failed for step input "{input_name}" - '
                f'expected type "{dagster_type.display_name}". '
                f"Description: {type_check.description}"
            ),
            metadata=type_check.metadata,
            dagster_type=dagster_type,
        )


def _type_check_output(
    step_context: StepExecutionContext,
    step_output_handle: StepOutputHandle,
    output: Any,
) -> Iterator[DagsterEvent]:
    check.inst_param(step_context, "step_context", StepExecutionContext)
    check.inst_param(output, "output", (Output, DynamicOutput))

    step_output = step_context.step.step_output_named(output.output_name)
    step_output_def = step_context.op_def.output_def_named(step_output.name)

    dagster_type = step_output_def.dagster_type
    type_check_context = step_context.for_type(dagster_type)
    op_label = step_context.describe_op()
    output_type = type(output.value)

    with user_code_error_boundary(
        DagsterTypeCheckError,
        lambda: (
            f'Error occurred while type-checking output "{output.output_name}" of {op_label}, with'
            f" Python type {output_type} and Dagster type {dagster_type.display_name}"
        ),
        log_manager=type_check_context.log,
    ):
        type_check = do_type_check(type_check_context, dagster_type, output.value)

    yield DagsterEvent.step_output_event(
        step_context=step_context,
        step_output_data=StepOutputData(
            step_output_handle=step_output_handle,
            type_check_data=TypeCheckData(
                success=type_check.success,
                label=step_output_handle.output_name,
                description=type_check.description if type_check else None,
                metadata=type_check.metadata if type_check else {},
            ),
            metadata=output.metadata,
        ),
    )

    if not type_check.success:
        raise DagsterTypeCheckDidNotPass(
            description=(
                f'Type check failed for step output "{output.output_name}" - '
                f'expected type "{dagster_type.display_name}". '
                f"Description: {type_check.description}"
            ),
            metadata=type_check.metadata,
            dagster_type=dagster_type,
        )


def core_dagster_event_sequence_for_step(
    step_context: StepExecutionContext,
) -> Iterator[DagsterEvent]:
    """Execute the step within the step_context argument given the in-memory
    events. This function yields a sequence of DagsterEvents, but without
    catching any exceptions that have bubbled up during the computation
    of the step.
    """
    check.inst_param(step_context, "step_context", StepExecutionContext)

    if step_context.previous_attempt_count > 0:
        yield DagsterEvent.step_restarted_event(step_context, step_context.previous_attempt_count)
    else:
        yield DagsterEvent.step_start_event(step_context)

    with (
        time_execution_scope() as timer_result,
        enter_execution_context(step_context) as compute_context,
    ):
        inputs = {}

        if step_context.is_sda_step:
            step_context.fetch_external_input_asset_version_info()

        for step_input in step_context.step.step_inputs:
            input_def = step_context.op_def.input_def_named(step_input.name)
            dagster_type = input_def.dagster_type

            if dagster_type.is_nothing:
                continue

            for event_or_input_value in step_input.source.load_input_object(
                step_context, input_def
            ):
                if isinstance(event_or_input_value, DagsterEvent):
                    yield event_or_input_value
                else:
                    check.invariant(step_input.name not in inputs)
                    inputs[step_input.name] = event_or_input_value

        for input_name, input_value in inputs.items():
            for evt in check.generator(
                _type_checked_event_sequence_for_input(step_context, input_name, input_value)
            ):
                yield evt

        # The core execution loop expects a compute generator in a specific format: a generator that
        # takes a context and dictionary of inputs as input, yields output events. If an op definition
        # was generated from the @op decorator, then compute_fn needs to be coerced
        # into this format. If the op definition was created directly, then it is expected that the
        # compute_fn is already in this format.
        if isinstance(step_context.op_def.compute_fn, DecoratedOpFunction):
            core_gen = create_op_compute_wrapper(step_context.op_def)
        else:
            core_gen = step_context.op_def.compute_fn

        user_event_sequence = execute_core_compute(
            step_context,
            inputs,
            core_gen,
            compute_context,
        )

        # It is important for this loop to be indented within the
        # timer block above in order for time to be recorded accurately.
        for user_event in _step_output_error_checked_user_event_sequence(
            step_context,
            _process_asset_results_to_events(step_context, user_event_sequence),
        ):
            if isinstance(user_event, DagsterEvent):
                yield user_event
            elif isinstance(user_event, (Output, DynamicOutput)):
                for evt in _type_check_and_store_output(step_context, user_event):
                    yield evt
            # for now, I'm ignoring AssetMaterializations yielded manually, but we might want
            # to do something with these in the above path eventually
            elif isinstance(user_event, AssetMaterialization):
                yield DagsterEvent.asset_materialization(step_context, user_event)
            elif isinstance(user_event, AssetObservation):
                yield DagsterEvent.asset_observation(step_context, user_event)
            elif isinstance(user_event, AssetCheckEvaluation):
                yield DagsterEvent.asset_check_evaluation(step_context, user_event)
            elif isinstance(user_event, ExpectationResult):
                yield DagsterEvent.step_expectation_result(step_context, user_event)
            else:
                check.failed(f"Unexpected event {user_event}, should have been caught earlier")

    yield DagsterEvent.step_success_event(
        step_context, StepSuccessData(duration_ms=timer_result.millis)
    )


def _type_check_and_store_output(
    step_context: StepExecutionContext, output: Union[DynamicOutput, Output]
) -> Iterator[DagsterEvent]:
    check.inst_param(step_context, "step_context", StepExecutionContext)
    check.inst_param(output, "output", (Output, DynamicOutput))

    mapping_key = output.mapping_key if isinstance(output, DynamicOutput) else None

    step_output_handle = StepOutputHandle(
        step_key=step_context.step.key, output_name=output.output_name, mapping_key=mapping_key
    )

    # If we are executing using the execute_in_process API, then we allow for the outputs of ops
    # to be directly captured to a dictionary after they are computed.
    if step_context.output_capture is not None:
        step_context.output_capture[step_output_handle] = output.value
    # capture output at the step level for threading the computed output values to hook context
    if (
        step_context.step_output_capture is not None
        and step_context.step_output_metadata_capture is not None
    ):
        step_context.step_output_capture[step_output_handle] = output.value
        step_context.step_output_metadata_capture[step_output_handle] = output.metadata

    yield from _type_check_output(step_context, step_output_handle, output)

    yield from _store_output(step_context, step_output_handle, output)


def _get_output_asset_events(
    asset_key: AssetKey,
    asset_partitions: Iterable[str],
    output: Union[Output, DynamicOutput],
    output_def: OutputDefinition,
    io_manager_metadata: Mapping[str, MetadataValue],
    step_context: StepExecutionContext,
    execution_type: AssetExecutionType,
) -> Iterator[Union[AssetMaterialization, AssetObservation]]:
    # Metadata scoped to all events for this asset.
    key_scoped_metadata = {**output.metadata, **io_manager_metadata}

    # Clear any cached record associated with this asset, since we are about to generate a new
    # materialization.
    step_context.wipe_input_asset_version_info(asset_key)
    tags: dict[str, str]
    if (
        execution_type == AssetExecutionType.MATERIALIZATION
        and step_context.is_external_input_asset_version_info_loaded
        and asset_key in step_context.job_def.asset_layer.executable_asset_keys
    ):
        assert isinstance(output, Output)
        code_version = _get_code_version(asset_key, step_context)
        input_provenance_data = _get_input_provenance_data(asset_key, step_context)
        cached_data_version = (
            step_context.get_data_version(asset_key)
            if step_context.has_data_version(asset_key)
            else None
        )
        user_provided_data_version = output.data_version or cached_data_version
        data_version = (
            compute_logical_data_version(
                code_version,
                {k: meta["data_version"] for k, meta in input_provenance_data.items()},
            )
            if user_provided_data_version is None
            else user_provided_data_version
        )
        tags = _build_data_version_tags(
            data_version,
            code_version,
            input_provenance_data,
            user_provided_data_version is not None,
        )
        if not step_context.has_data_version(asset_key):
            data_version = DataVersion(tags[DATA_VERSION_TAG])
            step_context.set_data_version(asset_key, data_version)
    elif execution_type == AssetExecutionType.OBSERVATION:
        assert isinstance(output, Output)
        tags = (
            _build_data_version_observation_tags(output.data_version) if output.data_version else {}
        )
    else:
        tags = {}

    all_tags = {**tags, **((output.tags if not isinstance(output, DynamicOutput) else None) or {})}

    backfill_id = step_context.get_tag(BACKFILL_ID_TAG)
    if backfill_id:
        tags[BACKFILL_ID_TAG] = backfill_id

    if execution_type == AssetExecutionType.MATERIALIZATION:
        event_class = AssetMaterialization
        event_class = AssetMaterialization
    elif execution_type == AssetExecutionType.OBSERVATION:
        event_class = AssetObservation
    else:
        check.failed(f"Unexpected asset execution type {execution_type}")

    unpartitioned_asset_metadata = step_context.get_asset_metadata(asset_key=asset_key)
    all_unpartitioned_asset_metadata = {
        **key_scoped_metadata,
        **(unpartitioned_asset_metadata or {}),
    }
    if asset_partitions:
        for partition in asset_partitions:
            with disable_dagster_warnings():
                partition_scoped_metadata = step_context.get_asset_metadata(
                    asset_key=asset_key, partition_key=partition
                )
                all_metadata_for_partitioned_event = {
                    **all_unpartitioned_asset_metadata,
                    **(partition_scoped_metadata or {}),
                }
                # copy the tags dictionary before setting the partition key tags. Otherwise
                # all asset materialization events will point to the same dictionary with the
                # partition key tags of the last partition processed.
                tags_for_event = {**all_tags}
                tags_for_event.update(
                    get_tags_from_multi_partition_key(partition)
                    if isinstance(partition, MultiPartitionKey)
                    else {}
                )

                yield event_class(
                    asset_key=asset_key,
                    partition=partition,
                    metadata=all_metadata_for_partitioned_event,
                    tags=tags_for_event,
                )
    else:
        with disable_dagster_warnings():
            yield event_class(
                asset_key=asset_key, metadata=all_unpartitioned_asset_metadata, tags=all_tags
            )


def _get_code_version(asset_key: AssetKey, step_context: StepExecutionContext) -> str:
    return (
        step_context.job_def.asset_layer.get(asset_key).code_version
        or step_context.dagster_run.run_id
    )


class _InputProvenanceData(TypedDict):
    data_version: DataVersion
    storage_id: Optional[int]


def _get_input_provenance_data(
    asset_key: AssetKey, step_context: StepExecutionContext
) -> Mapping[AssetKey, _InputProvenanceData]:
    input_provenance: dict[AssetKey, _InputProvenanceData] = {}
    deps = step_context.job_def.asset_layer.get(asset_key).parent_keys
    for key in deps:
        # For deps external to this step, this will retrieve the cached record that was stored prior
        # to step execution. For inputs internal to this step, it may trigger a query to retrieve
        # the most recent materialization record (it will retrieve a cached record if it's already
        # been asked for). For this to be correct, the output materializations for the step must be
        # generated in topological order -- we assume this.
        version_info = step_context.maybe_fetch_and_get_input_asset_version_info(key)

        # This can only happen for source assets that have never been observed.
        if version_info is None:
            storage_id = None
            data_version = DEFAULT_DATA_VERSION
        else:
            storage_id = version_info.storage_id
            data_version = version_info.data_version or DEFAULT_DATA_VERSION

        input_provenance[key] = {
            "data_version": data_version,
            "storage_id": storage_id,
        }
    return input_provenance


def _build_data_version_tags(
    data_version: DataVersion,
    code_version: str,
    input_provenance_data: Mapping[AssetKey, _InputProvenanceData],
    data_version_is_user_provided: bool,
) -> dict[str, str]:
    tags: dict[str, str] = {}
    tags[CODE_VERSION_TAG] = code_version
    for key, meta in input_provenance_data.items():
        tags[get_input_data_version_tag(key)] = meta["data_version"].value
        tags[get_input_event_pointer_tag(key)] = (
            str(meta["storage_id"]) if meta["storage_id"] else NULL_EVENT_POINTER
        )
    tags[DATA_VERSION_TAG] = data_version.value
    if data_version_is_user_provided:
        tags[DATA_VERSION_IS_USER_PROVIDED_TAG] = "true"
    return tags


def _build_data_version_observation_tags(data_version: DataVersion) -> dict[str, str]:
    return {
        DATA_VERSION_TAG: data_version.value,
        DATA_VERSION_IS_USER_PROVIDED_TAG: "true",
    }


def _store_output(
    step_context: StepExecutionContext,
    step_output_handle: StepOutputHandle,
    output: Union[Output, DynamicOutput],
) -> Iterator[DagsterEvent]:
    output_def = step_context.op_def.output_def_named(step_output_handle.output_name)
    output_manager = step_context.get_io_manager(step_output_handle)
    output_context = step_context.get_output_context(step_output_handle, output.metadata)

    manager_materializations = []
    manager_metadata: dict[str, MetadataValue] = {}

    # don't store asset check outputs, asset observation outputs, asset result outputs, or Nothing
    # type outputs
    step_output = step_context.step.step_output_named(step_output_handle.output_name)
    if (
        step_output.properties.asset_check_key
        or (step_context.output_observes_source_asset(step_output_handle.output_name))
        or isinstance(output, AssetResultOutput)
        or output_context.dagster_type.is_nothing
    ):
        yield from _log_materialization_or_observation_events_for_asset(
            step_context=step_context,
            output_context=output_context,
            output=output,
            output_def=output_def,
            manager_metadata={},
        )
    # otherwise invoke the I/O manager
    else:
        # output_manager.handle_output is either a generator function, or a normal function with or
        # without a return value. In the case that handle_output is a normal function, we need to
        # catch errors should they be raised before a return value. We can do this by wrapping
        # handle_output in a generator so that errors will be caught within iterate_with_context.
        if not inspect.isgeneratorfunction(output_manager.handle_output):

            def _gen_fn():
                gen_output = output_manager.handle_output(output_context, output.value)
                yield from output_context.consume_events()
                if gen_output:
                    yield gen_output

            handle_output_gen = _gen_fn()
        else:
            handle_output_gen = output_manager.handle_output(output_context, output.value)

        for elt in iterate_with_context(
            lambda: op_execution_error_boundary(
                DagsterExecutionHandleOutputError,
                msg_fn=lambda: f'Error occurred while handling output "{output_context.name}" of step "{step_context.step.key}":',
                step_context=step_context,
                step_key=step_context.step.key,
                output_name=output_context.name,
            ),
            handle_output_gen,
        ):
            for event in output_context.consume_events():
                yield event

            manager_metadata = {**manager_metadata, **output_context.consume_logged_metadata()}
            if isinstance(elt, DagsterEvent):
                yield elt
            elif isinstance(elt, AssetMaterialization):
                manager_materializations.append(elt)
            elif isinstance(elt, dict):  # should remove this?
                beta_warning("Yielding metadata from an IOManager's handle_output() function")
                manager_metadata = {**manager_metadata, **normalize_metadata(elt)}
            else:
                raise DagsterInvariantViolationError(
                    f"IO manager on output {output_def.name} has returned "
                    f"value {elt} of type {type(elt).__name__}. The return type can only be "
                    "one of AssetMaterialization, Dict[str, MetadataValue]."
                )

        for event in output_context.consume_events():
            yield event

        manager_metadata = {**manager_metadata, **output_context.consume_logged_metadata()}
        # do not alter explicitly created AssetMaterializations
        for mgr_materialization in manager_materializations:
            if mgr_materialization.metadata and manager_metadata:
                raise DagsterInvariantViolationError(
                    f"When handling output '{output_context.name}' of"
                    f" {output_context.op_def.node_type_str} '{output_context.op_def.name}', received a"
                    " materialization with metadata, while context.add_output_metadata was used within"
                    " the same call to handle_output. Due to potential conflicts, this is not allowed."
                    " Please specify metadata in one place within the `handle_output` function."
                )

            if manager_metadata:
                with disable_dagster_warnings():
                    materialization = AssetMaterialization(
                        asset_key=mgr_materialization.asset_key,
                        description=mgr_materialization.description,
                        metadata=manager_metadata,
                        partition=mgr_materialization.partition,
                    )
            else:
                materialization = mgr_materialization

            yield DagsterEvent.asset_materialization(step_context, materialization)

        yield from _log_materialization_or_observation_events_for_asset(
            step_context=step_context,
            output_context=output_context,
            output=output,
            output_def=output_def,
            manager_metadata=manager_metadata,
        )

        yield DagsterEvent.handled_output(
            step_context,
            output_name=step_output_handle.output_name,
            manager_key=output_def.io_manager_key,
            metadata=manager_metadata,
        )


def _log_materialization_or_observation_events_for_asset(
    step_context: StepExecutionContext,
    output_context: OutputContext,
    output: Union[Output, DynamicOutput],
    output_def: OutputDefinition,
    manager_metadata: Mapping[str, MetadataValue],
) -> Iterable[DagsterEvent]:
    # This is a temporary workaround to prevent duplicate observation events from external
    # observable assets that were auto-converted from source assets. These assets yield
    # observation events through the context in their body, and will continue to do so until we
    # can convert them to using ObserveResult, which requires a solution to partition-scoped
    # metadata and data version on output. We identify these auto-converted assets by looking
    # for OBSERVATION-type asset that have this special metadata key (added in
    # `wrap_source_asset_observe_fn_in_op_compute_fn`), which should only occur for these
    # auto-converted source assets. This can be removed when source asset observation functions
    # are converted to use ObserveResult.
    if SYSTEM_METADATA_KEY_SOURCE_ASSET_OBSERVATION in output.metadata:
        return

    asset_key = output_context.asset_key if output_context.has_asset_key else None
    partitions = output_context.asset_partition_keys if output_context.has_asset_partitions else []

    if asset_key:
        asset_layer = step_context.job_def.asset_layer
        assets_def = asset_layer.assets_def_for_node(step_context.node_handle)
        execution_type = check.not_none(assets_def).execution_type

        check.invariant(
            execution_type != AssetExecutionType.UNEXECUTABLE,
            "There should never be unexecutable assets here",
        )

        check.invariant(
            execution_type in {AssetExecutionType.MATERIALIZATION, AssetExecutionType.OBSERVATION},
            f"Unexpected asset execution type {execution_type}",
        )

        asset_events = list(
            _get_output_asset_events(
                asset_key,
                partitions,
                output,
                output_def,
                manager_metadata,
                step_context,
                execution_type,
            )
        )

        batch_id = generate_event_batch_id()
        last_index = len(asset_events) - 1
        for i, asset_event in enumerate(asset_events):
            batch_metadata = (
                DagsterEventBatchMetadata(batch_id, i == last_index) if partitions else None
            )
            yield _dagster_event_for_asset_event(step_context, asset_event, batch_metadata)


def _dagster_event_for_asset_event(
    step_context: StepExecutionContext,
    asset_event: Union[AssetMaterialization, AssetObservation],
    batch_metadata: Optional[DagsterEventBatchMetadata],
) -> DagsterEvent:
    if isinstance(asset_event, AssetMaterialization):
        return DagsterEvent.asset_materialization(step_context, asset_event, batch_metadata)
    else:  # observation
        return DagsterEvent.asset_observation(step_context, asset_event, batch_metadata)



import calendar
import datetime
import functools
import math
import re
from collections.abc import Iterator, Sequence
from typing import Optional, Union

import dagster._check as check
from dagster._core.definitions.partition import ScheduleType
from dagster._time import get_timezone
from dagster._vendored.croniter import croniter as _croniter
from dagster._vendored.dateutil.relativedelta import relativedelta
from dagster._vendored.dateutil.tz import datetime_ambiguous, datetime_exists

# Monthly schedules with 29-31 won't reliably run every month
MAX_DAY_OF_MONTH_WITH_GUARANTEED_MONTHLY_INTERVAL = 28

CRON_RANGES = ((0, 59), (0, 23), (1, 31), (1, 12), (0, 7), (0, 59))
CRON_STEP_SEARCH_REGEX = re.compile(r"^([^-]+)-([^-/]+)(/(\d+))?$")
INT_REGEX = re.compile(r"^\d+$")

# hours in which it is possible that a time might be ambigious or not exist due to daylight savings time
DAYLIGHT_SAVINGS_HOURS = {1, 2, 3}


class CroniterShim(_croniter):
    """Lightweight shim to enable caching certain values that may be calculated many times."""

    @classmethod
    @functools.lru_cache(maxsize=128)
    def expand(cls, *args, **kwargs):
        return super().expand(*args, **kwargs)


def _is_simple_cron(
    cron_expression: str,
    dt: datetime.datetime,
) -> bool:
    """This function is purely an optimization to see if the provided datetime is already on an obvious boundary
    of the common and easy to detect (daily at midnight and on the hour). The optimization is to avoid calling
    _find_schedule_time to find the next cron boundary.
    """
    if cron_expression == "0 0 * * *":
        return dt.hour == 0 and dt.minute == 0 and dt.second == 0 and dt.microsecond == 0

    if cron_expression == "0 * * * *":
        return dt.minute == 0 and dt.second == 0 and dt.microsecond == 0

    return False


def is_valid_cron_string(cron_string: str) -> bool:
    if not CroniterShim.is_valid(cron_string):
        return False

    # Croniter < 1.4 returns 2 items
    # Croniter >= 1.4 returns 3 items
    expanded, *_ = CroniterShim.expand(cron_string)

    # dagster only recognizes cron strings that resolve to 5 parts (e.g. not seconds resolution)
    if len(expanded) != 5:
        return False

    if len(expanded[3]) == 1 and expanded[3][0] == 2:  # February
        if len(expanded[2]) == 1 and expanded[2][0] in {30, 31}:  # 30th or 31st of February
            return False

    return True


def is_valid_cron_schedule(cron_schedule: Union[str, Sequence[str]]) -> bool:
    return (
        is_valid_cron_string(cron_schedule)
        if isinstance(cron_schedule, str)
        else len(cron_schedule) > 0
        and all(is_valid_cron_string(cron_string) for cron_string in cron_schedule)
    )


def cron_string_repeats_every_hour(cron_string: str) -> bool:
    """Returns if the given cron schedule repeats every hour."""
    cron_parts, nth_weekday_of_month, *_ = CroniterShim.expand(cron_string)
    return len(cron_parts[1]) == 1 and cron_parts[1][0] == "*"


def apply_fold_and_post_transition(date: datetime.datetime) -> datetime.datetime:
    date = apply_post_transition(date)
    return _apply_fold(date)


def _apply_fold(date: datetime.datetime) -> datetime.datetime:
    """For consistency, always choose the latter of the two possible times during a fall DST
    transition when there are two possibilities - match behavior described in the docs:
    https://docs.dagster.io/guides/automate/schedules/customizing-execution-timezone#execution-times-and-daylight-savings-time)

    Never call this with datetimes that could be non-existant. datetime_ambiguous will return true
    but folding them will leave them non-existant.
    """  # noqa: D415
    if date.fold == 0 and date.hour in DAYLIGHT_SAVINGS_HOURS and datetime_ambiguous(date):
        return date.replace(fold=1)
    return date


def apply_post_transition(
    date: datetime.datetime,
) -> datetime.datetime:
    if date.hour in DAYLIGHT_SAVINGS_HOURS and not datetime_exists(date):
        # If we fall on a non-existant time (e.g. between 2 and 3AM during a DST transition)
        # advance to the end of the window, which does exist - match behavior described in the docs:
        # https://docs.dagster.io/guides/automate/schedules/customizing-execution-timezone#execution-times-and-daylight-savings-time)

        # This assumes that all dst offsets are <= to an hour, which is true at time of writing.
        # The date passed to dst needs to be in DST to get the offset for the timezone.
        dst_offset = check.not_none(date.tzinfo).dst(date + datetime.timedelta(hours=1))

        # This assumes that all dst transitions happen on the hour, which is true at time of writing.
        # Rewind time to the start of the transition and then add the offset to get the first time out of the transition.
        start_dst = date.replace(minute=0, second=0, microsecond=0)
        return start_dst + check.not_none(dst_offset)
    return date


def _replace_date_fields(
    date: datetime.datetime,
    hour: int,
    minute: int,
    day: int,
):
    new_date = date.replace(
        day=day,
        hour=hour,
        minute=minute,
        second=0,
        microsecond=0,
    )
    return apply_fold_and_post_transition(new_date)


SECONDS_PER_MINUTE = 60
MINUTES_PER_HOUR = 60


def _find_hourly_schedule_time(
    minutes: Sequence[int],
    date: datetime.datetime,
    ascending: bool,
    already_on_boundary: bool,
) -> datetime.datetime:
    if ascending:
        # short-circuit if minutes and seconds are already correct
        if len(minutes) == 1 and (
            already_on_boundary
            or (date.minute == minutes[0] and date.second == 0 and date.microsecond == 0)
        ):
            # switch to utc so that timedelta behaves as expected instead of doing walltime math
            new_date = date.astimezone(datetime.timezone.utc) + datetime.timedelta(hours=1)
            new_date = new_date.astimezone(date.tzinfo)
            return new_date

        # clear microseconds
        new_timestamp = math.ceil(date.timestamp())
        # clear seconds
        new_timestamp = (
            new_timestamp
            + (SECONDS_PER_MINUTE - new_timestamp % SECONDS_PER_MINUTE) % SECONDS_PER_MINUTE
        )

        current_minute = (new_timestamp // SECONDS_PER_MINUTE) % SECONDS_PER_MINUTE

        final_timestamp = None

        for minute in minutes:
            new_timestamp_cand = new_timestamp + SECONDS_PER_MINUTE * (
                (minute - current_minute) % MINUTES_PER_HOUR
            )

            # move forward an hour if we haven't moved forwards yet
            if new_timestamp_cand <= date.timestamp():
                new_timestamp_cand = new_timestamp_cand + SECONDS_PER_MINUTE * MINUTES_PER_HOUR

            final_timestamp = (
                new_timestamp_cand
                if not final_timestamp
                else min(final_timestamp, new_timestamp_cand)
            )
    else:
        if len(minutes) == 1 and (
            already_on_boundary
            or (date.minute == minutes[0] and date.second == 0 and date.microsecond == 0)
        ):
            # switch to utc so that timedelta behaves as expected instead of doing walltime math
            new_date = date.astimezone(datetime.timezone.utc) - datetime.timedelta(hours=1)
            new_date = new_date.astimezone(date.tzinfo)
            return new_date

        # clear microseconds
        new_timestamp = math.floor(date.timestamp())
        # clear seconds
        new_timestamp = new_timestamp - new_timestamp % SECONDS_PER_MINUTE

        # move minutes back to correct place
        current_minute = (new_timestamp // SECONDS_PER_MINUTE) % SECONDS_PER_MINUTE

        final_timestamp = None

        for minute in minutes:
            new_timestamp_cand = new_timestamp - SECONDS_PER_MINUTE * (
                (current_minute - minute) % MINUTES_PER_HOUR
            )

            # move back an hour if we haven't moved backwards yet
            if new_timestamp_cand >= date.timestamp():
                new_timestamp_cand = new_timestamp_cand - SECONDS_PER_MINUTE * MINUTES_PER_HOUR

            final_timestamp = (
                new_timestamp_cand
                if not final_timestamp
                else max(final_timestamp, new_timestamp_cand)
            )

    return datetime.datetime.fromtimestamp(check.not_none(final_timestamp), tz=date.tzinfo)


def _find_daily_schedule_time(
    minute: int,
    hour: int,
    date: datetime.datetime,
    ascending: bool,
    already_on_boundary: bool,
) -> datetime.datetime:
    # First move to the correct time of day today (ignoring whether it is the correct day)
    if not already_on_boundary or (
        date.hour != hour or date.minute != minute or date.second != 0 or date.microsecond != 0
    ):
        new_time = _replace_date_fields(
            date,
            hour,
            minute,
            date.day,
        )
    else:
        new_time = date

    if ascending:
        if already_on_boundary or new_time.timestamp() <= date.timestamp():
            new_time = new_time + datetime.timedelta(days=1)
    else:
        if already_on_boundary or new_time.timestamp() >= date.timestamp():
            new_time = new_time - datetime.timedelta(days=1)

    # If the hour or minute has changed the schedule in cronstring,
    # double-check that it's still correct in case we crossed a DST boundary
    if new_time.hour != hour or new_time.minute != minute:
        new_time = _replace_date_fields(
            new_time,
            hour,
            minute,
            new_time.day,
        )
    return apply_fold_and_post_transition(new_time)


def _get_crontab_day_of_week(dt: datetime.datetime) -> int:
    weekday = dt.isoweekday()
    # crontab has 0-6, sunday - saturday
    # isoweekday is 1-7 monday - sunday
    return weekday if weekday <= 6 else 0


def _find_weekly_schedule_time(
    minute: int,
    hour: int,
    day_of_week: int,
    date: datetime.datetime,
    ascending: bool,
    already_on_boundary: bool,
) -> datetime.datetime:
    # first move to the correct time of day
    if not already_on_boundary:
        new_time = _replace_date_fields(
            date,
            hour,
            minute,
            date.day,
        )
        # Move to the correct day of the week
        current_day_of_week = _get_crontab_day_of_week(new_time)

        if day_of_week != current_day_of_week:
            if ascending:
                new_time = new_time + relativedelta(days=(day_of_week - current_day_of_week) % 7)
            else:
                new_time = new_time - relativedelta(days=(current_day_of_week - day_of_week) % 7)
    else:
        new_time = date

    # Make sure that we've actually moved in the correct direction, advance if we haven't
    if ascending:
        if already_on_boundary or new_time.timestamp() <= date.timestamp():
            new_time = new_time + relativedelta(weeks=1)
    else:
        if already_on_boundary or new_time.timestamp() >= date.timestamp():
            new_time = new_time - relativedelta(weeks=1)

    # If the hour or minute has from the schedule in cronstring,
    # double-check that it's still correct in case we crossed a DST boundary
    if new_time.hour != hour or new_time.minute != minute:
        new_time = _replace_date_fields(
            new_time,
            hour,
            minute,
            new_time.day,
        )
    return apply_fold_and_post_transition(new_time)


def _find_monthly_schedule_time(
    minute: int,
    hour: int,
    day: int,
    date: datetime.datetime,
    ascending: bool,
    already_on_boundary: bool,
) -> datetime.datetime:
    # First move to the correct day and time of day
    if not already_on_boundary:
        new_time = _replace_date_fields(
            date,
            check.not_none(hour),
            check.not_none(minute),
            check.not_none(day),
        )
    else:
        new_time = date

    if ascending:
        if already_on_boundary or new_time.timestamp() <= date.timestamp():
            new_time = new_time + relativedelta(months=1)
    else:
        if already_on_boundary or new_time.timestamp() >= date.timestamp():
            # Move back a month if needed
            new_time = new_time - relativedelta(months=1)

    # If the hour or minute has changed from the schedule in cronstring,
    # double-check that it's still correct in case we crossed a DST boundary
    if new_time.hour != hour or new_time.minute != minute:
        new_time = _replace_date_fields(
            new_time,
            check.not_none(hour),
            check.not_none(minute),
            check.not_none(day),
        )

    return apply_fold_and_post_transition(new_time)


def _find_schedule_time(
    minutes: Optional[Sequence[int]],
    hour: Optional[int],
    day_of_month: Optional[int],
    day_of_week: Optional[int],
    schedule_type: ScheduleType,
    date: datetime.datetime,
    ascending: bool,
    # lets us skip slow work to find the starting point if we know that
    # we are already on the boundary of the cron interval
    already_on_boundary: bool,
) -> datetime.datetime:
    if schedule_type == ScheduleType.HOURLY:
        return _find_hourly_schedule_time(
            check.not_none(minutes), date, ascending, already_on_boundary
        )
    elif schedule_type == ScheduleType.DAILY:
        minutes = check.not_none(minutes)
        check.invariant(len(minutes) == 1)
        return _find_daily_schedule_time(
            minutes[0],
            check.not_none(hour),
            date,
            ascending,
            already_on_boundary,
        )
    elif schedule_type == ScheduleType.WEEKLY:
        minutes = check.not_none(minutes)
        check.invariant(len(minutes) == 1)
        return _find_weekly_schedule_time(
            minutes[0],
            check.not_none(hour),
            check.not_none(day_of_week),
            date,
            ascending,
            already_on_boundary,
        )
    elif schedule_type == ScheduleType.MONTHLY:
        minutes = check.not_none(minutes)
        check.invariant(len(minutes) == 1)
        return _find_monthly_schedule_time(
            minutes[0],
            check.not_none(hour),
            check.not_none(day_of_month),
            date,
            ascending,
            already_on_boundary,
        )
    else:
        raise Exception(f"Unexpected schedule type {schedule_type}")


def _get_dates_to_consider_after_ambigious_time(
    cron_iter: CroniterShim,
    next_date: datetime.datetime,
    repeats_every_hour: bool,
    ascending: bool,
):
    # Return a list of all times that need to be considered when the next date returned by
    # croniter is ambigious (e.g. 2:30 AM during a fall DST transition). This is tricky because
    # we need to make sure that we are emitting times in the correct order, so we return a sorted
    # contiguous sequence of times that include all potentially ambigious times.
    post_transition_time = next_date.replace(fold=1)

    # Most schedules only need to consider the POST_TRANSITION time and can return here.
    if not repeats_every_hour:
        return [post_transition_time]

    # hourly schedules are more complicated - they'll continue firing once an hour no
    # matter what, including both PRE_TRANSITION and POST_TRANSITION times. So we need
    # to make sure that every time in the ambigious timerange has both its PRE_TRANSITION
    # and POST_TRANSITION times considered and returned.
    pre_transition_time = next_date.replace(fold=0)
    dates_to_consider = [post_transition_time, pre_transition_time]

    curr_pre_transition_time = pre_transition_time
    curr_post_transition_time = post_transition_time
    while True:
        if ascending:
            # Time always advances because get_next() is called, so we will eventually break
            # Stop once the current PRE_TRANSITION time exceeds the original POST_TRANSITION time
            # (so we know we have moved forward across the whole range)
            if curr_pre_transition_time.timestamp() >= post_transition_time.timestamp():
                break
            next_date = cron_iter.get_next(datetime.datetime)
        else:
            # Time always decreases because get_prev() is called, so we will eventually break
            # Stop once the current POST_TRANSITION time has gone past the original PRE_TRANSITION
            # time (so we know we have moved backward across the whole range)
            if curr_post_transition_time.timestamp() <= pre_transition_time.timestamp():
                break
            next_date = cron_iter.get_prev(datetime.datetime)

        # Make sure we add both the PRE_TRANSITION and POST_TRANSITION times to the
        # list of dates to consider so every time emitted by the
        # croniter instance is considered and returned from the calling iterator
        curr_pre_transition_time = datetime.datetime(
            year=next_date.year,
            month=next_date.month,
            day=next_date.day,
            hour=next_date.hour,
            minute=next_date.minute,
            second=next_date.second,
            microsecond=next_date.microsecond,
            fold=0,
            tzinfo=post_transition_time.tzinfo,
        )
        dates_to_consider.append(curr_pre_transition_time)
        curr_post_transition_time = datetime.datetime(
            year=next_date.year,
            month=next_date.month,
            day=next_date.day,
            hour=next_date.hour,
            minute=next_date.minute,
            second=next_date.second,
            microsecond=next_date.microsecond,
            fold=1,
            tzinfo=post_transition_time.tzinfo,
        )
        dates_to_consider.append(curr_post_transition_time)

    return sorted(dates_to_consider, key=lambda d: d.timestamp())


def _timezone_aware_cron_iter(
    cron_string, start_datetime: datetime.datetime, ascending: bool
) -> Iterator[datetime.datetime]:
    """Use croniter to determine the next timestamp matching the passed in cron string
    that is past the passed in UTC timestamp. croniter can only be trusted to compute
    non-timezone aware cron intervals, so we first figure out the time corresponding to the
    passed in timestamp without taking any timezones into account, use croniter to
    determine the next time that matches the cron string, translate that back into the passed in
    timezone, and repeat, returning the first time that is later than the passed in timestamp.
    """
    # Create a naive (timezone-free) datetime to pass into croniter
    naive_time = datetime.datetime(
        year=start_datetime.year,
        month=start_datetime.month,
        day=start_datetime.day,
        hour=start_datetime.hour,
        minute=start_datetime.minute,
        second=start_datetime.second,
        microsecond=start_datetime.microsecond,
    )

    # Go back an hour to ensure that we consider the full set of possible candidates (otherwise
    # we might fail to properly consider a time that happens twice during a fall DST transition).
    # 1 hour is sufficient because that's the maximum amount of time that can be offset during a
    # DST transition.
    if ascending:
        naive_time = naive_time - datetime.timedelta(hours=1)
    else:
        naive_time = naive_time + datetime.timedelta(hours=1)

    cron_iter = CroniterShim(cron_string, naive_time)

    # hourly schedules handle DST transitions differently: they skip times that don't exist
    # entirely and just move on to the next matching time (instead of returning
    # the end time of the non-existant interval), and when there are two times that match the cron
    # string, they return both instead of picking the latter time.
    repeats_every_hour = cron_string_repeats_every_hour(cron_string)

    # Chronological order of dates to return
    dates_to_consider = []
    start_timestamp = start_datetime.timestamp()
    while True:
        # Work through everything currently in dates_to_consider
        if ascending:
            for next_date_with_tz in dates_to_consider:
                next_timestamp = next_date_with_tz.timestamp()
                if next_timestamp > start_timestamp:
                    start_timestamp = next_timestamp
                    yield next_date_with_tz
        else:
            for next_date_with_tz in reversed(dates_to_consider):
                next_timestamp = next_date_with_tz.timestamp()
                if next_timestamp < start_timestamp:
                    start_timestamp = next_timestamp
                    yield next_date_with_tz

        # Clear the list and generate new candidates using croniter
        dates_to_consider = []

        if ascending:
            next_date = cron_iter.get_next(datetime.datetime)
        else:
            next_date = cron_iter.get_prev(datetime.datetime)
        next_date_with_tz = datetime.datetime(
            year=next_date.year,
            month=next_date.month,
            day=next_date.day,
            hour=next_date.hour,
            minute=next_date.minute,
            second=next_date.second,
            microsecond=next_date.microsecond,
            tzinfo=start_datetime.tzinfo,
        )

        dates_to_consider = [next_date_with_tz]
        if not datetime_exists(next_date_with_tz):
            if repeats_every_hour:
                # hourly schedules just move on to the next time
                dates_to_consider = []
            else:
                # other schedules advance to the time at the end of the interval (so that e.g.
                # a daily schedule doesn't miss an entire day)
                dates_to_consider = [apply_post_transition(next_date_with_tz)]
        elif datetime_ambiguous(next_date_with_tz):
            dates_to_consider = _get_dates_to_consider_after_ambigious_time(
                cron_iter=cron_iter,
                next_date=next_date_with_tz,
                repeats_every_hour=repeats_every_hour,
                ascending=ascending,
            )


def _has_out_of_range_cron_interval_str(cron_string: str):
    assert CroniterShim.is_valid(cron_string)
    try:
        for i, cron_part in enumerate(cron_string.lower().split()):
            expr_parts = cron_part.split(",")
            while len(expr_parts) > 0:
                expr = expr_parts.pop()
                t = re.sub(
                    r"^\*(\/.+)$",
                    r"%d-%d\1" % (CRON_RANGES[i][0], CRON_RANGES[i][1]),  # noqa: UP031
                    str(expr),
                )
                m = CRON_STEP_SEARCH_REGEX.search(t)
                if not m:
                    # try normalizing "{start}/{step}" to "{start}-{max}/{step}".
                    t = re.sub(r"^(.+)\/(.+)$", r"\1-%d/\2" % (CRON_RANGES[i][1]), str(expr))  # noqa: UP031
                    m = CRON_STEP_SEARCH_REGEX.search(t)
                if m:
                    (low, high, step) = m.group(1), m.group(2), m.group(4) or 1
                    if i == 2 and high == "l":
                        high = "31"
                    if not INT_REGEX.search(low) or not INT_REGEX.search(high):
                        continue
                    low, high, step = map(int, [low, high, step])
                    if step > high:
                        return True
    except:
        pass
    return False


def has_out_of_range_cron_interval(cron_schedule: Union[str, Sequence[str]]):
    """Utility function to detect cron schedules like '*/90 * * * *', which are valid cron schedules
    but which evaluate to once every hour, not once every 90 minutes as might be expected.  This is
    useful to detect so that we can issue warnings or some other kind of feedback to the user.  This
    function does not detect cases where the step does not divide cleanly in the range, which is
    another case that might cause some surprising behavior (e.g. '*/7 * * * *').
    """
    return (
        _has_out_of_range_cron_interval_str(cron_schedule)
        if isinstance(cron_schedule, str)
        else any(_has_out_of_range_cron_interval_str(s) for s in cron_schedule)
    )


def cron_string_iterator(
    start_timestamp: float,
    cron_string: str,
    execution_timezone: Optional[str],
    ascending: bool = True,
    start_offset: int = 0,
) -> Iterator[datetime.datetime]:
    """Generator of datetimes >= start_timestamp for the given cron string."""
    # leap day special casing
    if cron_string.endswith(" 29 2 *"):
        min_hour, _ = cron_string.split(" 29 2 *")
        day_before = f"{min_hour} 28 2 *"
        # run the iterator for Feb 28th
        for dt in cron_string_iterator(
            start_timestamp=start_timestamp,
            cron_string=day_before,
            execution_timezone=execution_timezone,
            ascending=ascending,
            start_offset=start_offset,
        ):
            # only return on leap years
            if calendar.isleap(dt.year):
                # shift 28th back to 29th
                shifted_dt = dt + datetime.timedelta(days=1)
                yield shifted_dt
        return
    execution_timezone = execution_timezone or "UTC"

    # Croniter < 1.4 returns 2 items
    # Croniter >= 1.4 returns 3 items
    cron_parts, nth_weekday_of_month, *_ = CroniterShim.expand(cron_string)

    is_numeric = [len(part) == 1 and isinstance(part[0], int) for part in cron_parts]
    is_wildcard = [len(part) == 1 and part[0] == "*" for part in cron_parts]

    all_numeric_minutes = len(cron_parts[0]) > 0 and all(
        cron_part != "*" for cron_part in cron_parts[0]
    )

    known_schedule_type: Optional[ScheduleType] = None

    expected_hour = None
    expected_minutes = None
    expected_day = None
    expected_day_of_week = None

    # Special-case common intervals (hourly/daily/weekly/monthly) since croniter iteration can be
    # much slower and has correctness issues on DST boundaries
    if not nth_weekday_of_month:
        if (
            all(is_numeric[0:3])
            and all(is_wildcard[3:])
            and cron_parts[2][0] <= MAX_DAY_OF_MONTH_WITH_GUARANTEED_MONTHLY_INTERVAL  # pyright: ignore[reportOperatorIssue]
        ):  # monthly
            known_schedule_type = ScheduleType.MONTHLY
        elif all(is_numeric[0:2]) and is_numeric[4] and all(is_wildcard[2:4]):  # weekly
            known_schedule_type = ScheduleType.WEEKLY
        elif all(is_numeric[0:2]) and all(is_wildcard[2:]):  # daily
            known_schedule_type = ScheduleType.DAILY
        elif all_numeric_minutes and all(is_wildcard[1:]):  # hourly
            known_schedule_type = ScheduleType.HOURLY

    if is_numeric[1]:
        expected_hour = cron_parts[1][0]

    if all_numeric_minutes:
        expected_minutes = [cron_part for cron_part in cron_parts[0]]

    if is_numeric[2]:
        expected_day = cron_parts[2][0]

    if is_numeric[4]:
        expected_day_of_week = cron_parts[4][0]

    if known_schedule_type:
        start_datetime = datetime.datetime.fromtimestamp(
            start_timestamp, tz=get_timezone(execution_timezone)
        )

        if start_offset == 0 and _is_simple_cron(cron_string, start_datetime):
            # In simple cases, where you're already on a cron boundary, the below logic is unnecessary
            # and slow
            next_date = start_datetime
            # This is already on a cron boundary, so yield it
            yield start_datetime
        else:
            next_date = _find_schedule_time(
                expected_minutes,  # pyright: ignore[reportArgumentType]
                expected_hour,  # pyright: ignore[reportArgumentType]
                expected_day,  # pyright: ignore[reportArgumentType]
                expected_day_of_week,  # pyright: ignore[reportArgumentType]
                known_schedule_type,
                start_datetime,
                ascending=not ascending,  # Going in the reverse direction
                already_on_boundary=False,
            )
            check.invariant(start_offset <= 0)
            for _ in range(-start_offset):
                next_date = _find_schedule_time(
                    expected_minutes,  # pyright: ignore[reportArgumentType]
                    expected_hour,  # pyright: ignore[reportArgumentType]
                    expected_day,  # pyright: ignore[reportArgumentType]
                    expected_day_of_week,  # pyright: ignore[reportArgumentType]
                    known_schedule_type,
                    next_date,
                    ascending=not ascending,  # Going in the reverse direction
                    already_on_boundary=True,
                )

        while True:
            next_date = _find_schedule_time(
                expected_minutes,  # pyright: ignore[reportArgumentType]
                expected_hour,  # pyright: ignore[reportArgumentType]
                expected_day,  # pyright: ignore[reportArgumentType]
                expected_day_of_week,  # pyright: ignore[reportArgumentType]
                known_schedule_type,
                next_date,
                ascending=ascending,
                already_on_boundary=True,
            )

            if start_offset == 0:
                if ascending:
                    # Guard against _find_schedule_time returning unexpected results
                    check.invariant(next_date.timestamp() >= start_timestamp)
                else:
                    check.invariant(next_date.timestamp() <= start_timestamp)

            yield next_date
    else:
        yield from _croniter_string_iterator(
            start_timestamp, cron_string, execution_timezone, ascending, start_offset
        )


def _croniter_string_iterator(
    start_timestamp: float,
    cron_string: str,
    timezone_str: str,
    ascending: bool = True,
    start_offset: int = 0,
):
    start_datetime = datetime.datetime.fromtimestamp(start_timestamp, get_timezone(timezone_str))
    reverse_cron = _timezone_aware_cron_iter(cron_string, start_datetime, ascending=not ascending)
    next_date = None
    check.invariant(start_offset <= 0)
    for _ in range(-start_offset + 1):
        next_date = next(reverse_cron)
    next_date = check.not_none(next_date).astimezone(start_datetime.tzinfo)
    forward_cron = _timezone_aware_cron_iter(cron_string, next_date, ascending=ascending)
    while True:
        next_date = next(forward_cron)

        if start_offset == 0:
            if ascending:
                # Guard against _find_schedule_time returning unexpected results
                check.invariant(next_date.timestamp() >= start_timestamp)
            else:
                check.invariant(next_date.timestamp() <= start_timestamp)

        yield next_date


def reverse_cron_string_iterator(
    end_timestamp: float,
    cron_string: str,
    execution_timezone: Optional[str],
) -> Iterator[datetime.datetime]:
    yield from cron_string_iterator(end_timestamp, cron_string, execution_timezone, ascending=False)


def schedule_execution_time_iterator(
    start_timestamp: float,
    cron_schedule: Union[str, Sequence[str]],
    execution_timezone: Optional[str],
    ascending: bool = True,
) -> Iterator[datetime.datetime]:
    """Generator of execution datetimes >= start_timestamp for the given schedule.

    Here cron_schedule is either a cron string or a sequence of cron strings. In the latter case,
    the next execution datetime is obtained by computing the next cron datetime
    after the current execution datetime for each cron string in the sequence, and then choosing
    the earliest among them.
    """
    check.invariant(
        is_valid_cron_schedule(cron_schedule), desc=f"{cron_schedule} must be a valid cron schedule"
    )

    if isinstance(cron_schedule, str):
        yield from (
            cron_string_iterator(start_timestamp, cron_schedule, execution_timezone)
            if ascending
            else reverse_cron_string_iterator(start_timestamp, cron_schedule, execution_timezone)
        )
    else:
        iterators = [
            (
                cron_string_iterator(start_timestamp, cron_string, execution_timezone)
                if ascending
                else reverse_cron_string_iterator(start_timestamp, cron_string, execution_timezone)
            )
            for cron_string in cron_schedule
        ]
        next_dates = [next(it) for it in iterators]
        while True:
            # Choose earliest out of all subsequent datetimes.
            earliest_next_date = min(next_dates)
            yield earliest_next_date
            # Increment all iterators that generated the earliest subsequent datetime.
            for i, next_date in enumerate(next_dates):
                if next_date == earliest_next_date:
                    next_dates[i] = next(iterators[i])


def get_latest_completed_cron_tick(
    cron_string: str,
    current_time: datetime.datetime,
    timezone: Optional[str],
) -> datetime.datetime:
    cron_iter = reverse_cron_string_iterator(
        end_timestamp=current_time.timestamp(),
        cron_string=cron_string,
        execution_timezone=timezone,
    )
    return next(cron_iter)


def get_next_cron_tick(
    cron_string: str,
    current_time: datetime.datetime,
    timezone: Optional[str],
) -> datetime.datetime:
    cron_iter = cron_string_iterator(
        start_timestamp=current_time.timestamp(),
        cron_string=cron_string,
        execution_timezone=timezone,
    )
    return next(cron_iter)



"""This file contains a set of utilities for dealing with test
environments where we want to go back and forth between
abstractions that reside in user process (e.g. definitions and
reconstructables) and abstractions that reside in host processes
(e.g. handles and externals).

These should only be invoked from contexts where we know this
to be the case.
"""

from collections.abc import Iterable
from typing import AbstractSet, Optional  # noqa: UP035

import dagster._check as check
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.reconstruct import ReconstructableJob, ReconstructableRepository
from dagster._core.origin import JobPythonOrigin, RepositoryPythonOrigin
from dagster._core.remote_representation import RemoteJob
from dagster._core.remote_representation.external_data import JobDataSnap
from dagster._core.remote_representation.handle import RepositoryHandle


def recon_job_from_origin(origin: JobPythonOrigin) -> ReconstructableJob:
    check.inst_param(origin, "origin", JobPythonOrigin)
    recon_repo = recon_repository_from_origin(origin.repository_origin)
    return recon_repo.get_reconstructable_job(origin.job_name)


def recon_repository_from_origin(origin: RepositoryPythonOrigin) -> "ReconstructableRepository":
    check.inst_param(origin, "origin", RepositoryPythonOrigin)
    return ReconstructableRepository(
        origin.code_pointer,
        origin.container_image,
        origin.executable_path,
        origin.entry_point,
        origin.container_context,
    )


def remote_job_from_recon_job(
    recon_job: ReconstructableJob,
    op_selection: Optional[Iterable[str]],
    repository_handle: RepositoryHandle,
    asset_selection: Optional[AbstractSet[AssetKey]] = None,
) -> RemoteJob:
    if op_selection or asset_selection:
        sub_recon_job = recon_job.get_subset(
            op_selection=op_selection, asset_selection=asset_selection
        )
        job_def = sub_recon_job.get_definition()
    else:
        job_def = recon_job.get_definition()

    return RemoteJob(
        JobDataSnap.from_job_def(job_def, include_parent_snapshot=True),
        repository_handle=repository_handle,
    )
